# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import Any
from typing import cast
from typing import Generic
from typing import TypeVar

import datetime as dt
import logging

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common.const import AvatarSize
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import ResourceContact
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.util.datetime import utc_now
from gajim.common.util.muc import get_groupchat_name
from gajim.common.util.user_strings import get_uf_relative_time
from gajim.plugins.manifest import PluginManifest
from gajim.plugins.repository import PluginRepository

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string

if typing.TYPE_CHECKING:
    from gajim.gtk.activity_page import ActivityPage


log = logging.getLogger("gajim.gtk.activity_list")

E = TypeVar("E", bound=events.ApplicationEvent | None)

EventT = (
    events.MucInvitation
    | events.MucDecline
    | events.SubscribePresenceReceived
    | events.UnsubscribedPresenceReceived
    | events.GajimUpdateAvailable
    | events.AllowGajimUpdateCheck
    | events.ReactionUpdated
)


class ActivityListView(Gtk.ListView, SignalManager, EventHelper):
    __gtype_name__ = "ActivityListView"
    __gsignals__ = {
        "unselected": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    unread_count = GObject.Property(type=int, default=0)

    def __init__(self) -> None:
        Gtk.ListView.__init__(self)
        SignalManager.__init__(self)
        EventHelper.__init__(self)

        self.add_css_class("activity-list-view")

        self._model: Gio.ListStore[ActivityListItem[Any]] = Gio.ListStore(
            item_type=ActivityListItem
        )

        factory = Gtk.SignalListItemFactory()
        self._connect(factory, "setup", self._on_factory_setup)
        self._connect(factory, "bind", self._on_factory_bind)
        self._connect(factory, "unbind", self._on_factory_unbind)
        self.set_factory(factory)

        # TODO
        # self._chat_filters = ChatFilters()
        self._scroll_id = None
        self._search_string_list: list[str] = []
        self._search_entry: Gtk.SearchEntry | None = None
        self._current_filter_text = ""

        self._timer_id = GLib.timeout_add_seconds(60, self._update_time)

        self._custom_filter = Gtk.CustomFilter.new(self._filter_func)

        self._filter_model = Gtk.FilterListModel(
            model=self._model, filter=self._custom_filter
        )

        self._selection_model = Gtk.SingleSelection(
            model=self._filter_model,
            autoselect=False,
            can_unselect=True,
        )
        self._connect(
            self._selection_model, "notify::selected-item", self._on_selection_changed
        )
        self._connect(self, "activate", self._on_activity_item_activate)
        self._connect(self, "unselected", self._on_activity_item_unselected)

        self.set_model(self._selection_model)

        self.register_events(
            [
                ("muc-invitation", ged.GUI1, self._on_event),
                ("muc-decline", ged.GUI1, self._on_event),
                ("subscription-request", ged.GUI1, self._on_event),
                ("unsubscribed-presence-received", ged.GUI1, self._on_event),
                ("gajim-update-available", ged.GUI1, self._on_event),
                ("allow-gajim-update-check", ged.GUI1, self._on_event),
                ("reaction-updated", ged.GUI1, self._on_event),
                ("account-disabled", ged.GUI1, self._on_account_disabled),
                ("timezone-changed", ged.GUI2, self._on_timezone_changed),
                ("openpgp-event", ged.GUI2, self._on_openpgp_event),
            ]
        )

        app.plugin_repository.connect(
            "plugin-updates-available", self._on_plugin_updates_available
        )
        app.plugin_repository.connect(
            "auto-update-finished", self._on_plugin_auto_update_finished
        )

        self._event_item_map = {
            events.MucInvitation: MucInvitation,
            events.MucDecline: MucInvitationDeclined,
            events.SubscribePresenceReceived: Subscribe,
            events.UnsubscribedPresenceReceived: Unsubscribed,
            events.GajimUpdateAvailable: GajimUpdate,
            events.AllowGajimUpdateCheck: GajimUpdatePermission,
            events.ReactionUpdated: Reaction,
        }

    def do_unroot(self) -> None:
        # The filter func needs to be unset before calling do_unroot (see #12213)
        self._custom_filter.set_filter_func(None)
        Gtk.ListView.do_unroot(self)
        self._disconnect_all()
        app.plugin_repository.disconnect_all_from_obj(self)

        GLib.source_remove(self._timer_id)
        app.check_finalize(self._model)
        app.check_finalize(self._filter_model)
        app.check_finalize(self._selection_model)
        app.check_finalize(self._custom_filter)
        del self._model
        del self._filter_model
        del self._selection_model
        del self._search_entry
        del self._custom_filter
        app.check_finalize(self)

    def set_page(self, page: ActivityPage) -> None:
        self._activity_page = page
        self._activity_page.connect("page-removed", self._on_activity_page_removed)

    def _on_activity_page_removed(
        self, page: ActivityPage, item: ActivityListItemT
    ) -> None:
        self._remove(item)

    def select_with_context_id(self, context_id: str) -> None:
        for position, item in enumerate(self._model):
            if item.context_id == context_id:
                self._selection_model.set_selected(position)
                return

    def mark_all_items_as_read(self) -> None:
        for item in self._model:
            item.read = True

        self.set_property("unread-count", 0)

    def mark_items_as_read(
        self,
        item_type: type[ActivityListItemT],
        account: str | None = None,
        jid: JID | None = None,
    ) -> None:
        for item in self._model:
            if not isinstance(item, item_type):
                continue

            if account is not None and item.account != account:
                continue

            if jid is not None and item.jid != jid:
                continue

            if not item.read:
                self._decrease_unread_count()
                item.read = True

    def unselect(self) -> None:
        self._selection_model.unselect_all()

    def _add(self, item: ActivityListItemT, *, prepend: bool = True) -> None:
        if item.unique:
            self._remove_same_items(item)

        if prepend:
            self._model.insert(0, item)
        else:
            self._model.append(item)

        if not item.read:
            self._increase_unread_count()

    def _remove_same_items(self, new_item: ActivityListItemT) -> None:
        for i in reversed(range(len(self._model))):
            old_item = self._model.get_item(i)
            assert old_item is not None

            if (
                isinstance(old_item, type(new_item))
                and old_item.account == new_item.account
            ):
                if not old_item.read:
                    self._decrease_unread_count()
                self._model.remove(i)

    def remove_by_type(
        self, item_type: type[ActivityListItemT], account: str | None = None
    ) -> None:
        self.unselect()
        self._remove_by_type(item_type, account)

    def _remove_by_type(
        self, item_type: type[ActivityListItemT], account: str | None = None
    ) -> None:
        for i in reversed(range(len(self._model))):
            item = self._model.get_item(i)
            assert item is not None

            if isinstance(item, item_type):
                if account is not None and item.account != account:
                    continue

                if not item.read:
                    self._decrease_unread_count()
                self._model.remove(i)

    def _remove(self, item: ActivityListItemT) -> None:
        found, position = self._model.find(item)
        if not found:
            return

        self._model.remove(position)
        if not item.read:
            self._decrease_unread_count()

    def _update_time(self) -> bool:
        now = utc_now()
        for item in self._model:
            delta = now - item.timestamp
            if delta > dt.timedelta(days=8):
                # Items older than 8 days show an absolute timestamp
                # so updating makes no sense anymore
                break

            item.update_time()

        return GLib.SOURCE_CONTINUE

    def _on_activity_item_activate(
        self, listview: ActivityListView, position: int
    ) -> None:
        item = listview.get_listitem(position)
        item.activated()
        self._activity_page.process_row_activated(item)

    def _on_activity_item_unselected(self, _listview: ActivityListView) -> None:
        self._activity_page.show_default_page()

    def _increase_unread_count(self) -> None:
        self.set_property("unread-count", self.unread_count + 1)

    def _decrease_unread_count(self) -> None:
        self.set_property("unread-count", self.unread_count - 1)

    def _on_selection_changed(
        self, selection_model: Gtk.SingleSelection, *args: Any
    ) -> None:
        item = selection_model.get_selected_item()
        if item is None:
            self.emit("unselected")
            return

        assert isinstance(item, ActivityListItem)
        if not item.read:
            item.read = True
            self._decrease_unread_count()
            if item.context_id:
                app.ged.raise_event(events.NotificationWithdrawn(item.context_id))

        position = selection_model.get_selected()
        self.activate_action("list.activate-item", GLib.Variant("u", position))

    def set_search_entry(self, entry: Gtk.SearchEntry) -> None:
        self._search_entry = entry
        self._connect(entry, "search-changed", self._on_search_changed)

    def _on_search_changed(self, entry: Gtk.SearchEntry) -> None:
        self._current_filter_text = entry.get_text().lower()
        self._custom_filter.changed(Gtk.FilterChange.DIFFERENT)

    @staticmethod
    def _on_factory_setup(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        list_item.set_child(ActivityViewItem())

    @staticmethod
    def _on_factory_bind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        view_item = cast(ActivityViewItem, list_item.get_child())
        obj = cast(ActivityListItem[EventT], list_item.get_item())
        view_item.bind(obj)

    @staticmethod
    def _on_factory_unbind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        view_item = cast(ActivityViewItem, list_item.get_child())
        view_item.unbind()

    def get_listitem(self, position: int) -> ActivityListItemT:
        return cast(ActivityListItemT, self._filter_model.get_item(position))

    def _filter_func(self, item: ActivityListItem[EventT]) -> bool:
        if not self._current_filter_text:
            return True

        return self._current_filter_text in item.search_text

    def _on_event(self, event: EventT) -> None:
        list_item_cls = self._event_item_map[type(event)]
        if not list_item_cls.can_create(event):  # type: ignore
            return

        item = list_item_cls.from_event(event)  # type: ignore
        self._add(item)

        if item.should_notify():
            app.ged.raise_event(
                events.Notification(
                    context_id=item.context_id,
                    account=item.account,
                    jid=None,
                    type=event.name,
                    title=item.title,
                    text=item.subject,
                )
            )

    def _on_plugin_updates_available(
        self,
        _repository: PluginRepository,
        _signal_name: str,
        manifests: list[PluginManifest],
    ) -> None:
        item = GajimPluginUpdate.from_event(
            events.PluginUpdatesAvailable(manifests=manifests)
        )
        self._add(item)

    def _on_plugin_auto_update_finished(
        self, _repository: PluginRepository, _signal_name: str
    ) -> None:
        self._remove_by_type(GajimPluginUpdate)
        item = GajimPluginUpdateFinished.from_event()
        self._add(item)

    def _on_account_disabled(self, event: events.AccountDisabled) -> None:
        count = self._model.get_n_items()
        for pos in range(count - 1, -1, -1):
            item = cast(ActivityListItem[Any], self._model.get_item(pos))
            if item.account == event.account:
                self._model.remove(pos)

    def _on_timezone_changed(self, event: events.TimezoneChanged) -> None:
        self._add(TimezoneChanged.from_event(event))

    def _on_openpgp_event(self, event: events.OpenPGPEvent) -> None:
        self._add(OpenPGPEvent.from_event(event))


class ActivityListItem(Generic[E], GObject.Object):
    __gtype_name__ = "ActivityListItem"

    context_id: str = GObject.Property(type=str)  # type: ignore
    account: str = GObject.Property(type=str)  # type: ignore
    jid: JID = GObject.Property(type=str)  # type: ignore
    account_visible: bool = GObject.Property(type=bool, default=False)  # type: ignore
    activity_type: int = GObject.Property(type=int)  # type: ignore
    activity_type_icon: str = GObject.Property(type=str)  # type: ignore
    avatar: Gdk.Paintable = GObject.Property(type=Gdk.Paintable)  # type: ignore
    timestamp: dt.datetime = GObject.Property(type=object)  # type: ignore
    title: str = GObject.Property(type=str)  # type: ignore
    subject: str = GObject.Property(type=str)  # type: ignore
    read: bool = GObject.Property(type=bool, default=False)  # type: ignore
    search_text: str = GObject.Property(type=str)  # type: ignore
    event: E = GObject.Property(type=object)  # type: ignore
    state = GObject.Property(type=object)
    unique: bool = GObject.Property(type=bool, default=False)  # type: ignore

    def __init__(
        self,
        context_id: str,
        account: str,
        jid: JID | None,
        account_visible: bool,
        activity_type: int,
        activity_type_icon: str,
        avatar: Gdk.Paintable,
        title: str,
        timestamp: dt.datetime,
        subject: str,
        read: bool,
        event: E,
        unique: bool = False,
    ) -> None:
        self._timestamp_string = get_uf_relative_time(timestamp)

        super().__init__(
            context_id=context_id,
            account=account,
            jid=jid,
            account_visible=account_visible,
            activity_type=activity_type,
            activity_type_icon=activity_type_icon,
            avatar=avatar,
            title=title,
            timestamp=timestamp,
            subject=subject,
            read=read,
            search_text=f"{title} {subject}",
            event=event,
            state={},
            unique=unique,
        )

    @GObject.Property(type=str, flags=GObject.ParamFlags.READABLE)
    def timestamp_string(self) -> str:
        return self._timestamp_string

    def update_time(self) -> None:
        self._timestamp_string = get_uf_relative_time(self.timestamp)
        self.notify("timestamp-string")

    def get_event(self) -> E:
        return self.event

    def activated(self) -> None:
        return

    def should_notify(self) -> bool:
        return False

    @staticmethod
    def can_create(event: E) -> bool:
        return True

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.account} - {self.jid}"


@Gtk.Template(string=get_ui_string("activity_list_row.ui"))
class ActivityViewItem(Gtk.Grid, SignalManager):
    __gtype_name__ = "ActivityViewItem"

    _account_identifier: Gtk.Box = Gtk.Template.Child()
    _avatar_image: Gtk.Image = Gtk.Template.Child()
    _activity_type_image: Gtk.Image = Gtk.Template.Child()
    _activity_title_label: Gtk.Label = Gtk.Template.Child()
    _timestamp_label: Gtk.Label = Gtk.Template.Child()
    _activity_subject_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Grid.__init__(self)
        SignalManager.__init__(self)

        self._account = ""
        self._jid = None
        self._account_css_class = ""
        self._read = False
        self.__bindings: list[GObject.Binding] = []

    @GObject.Property(type=str)
    def account(self) -> str:  # type: ignore
        return self._account

    @account.setter
    def account(self, account: str) -> None:
        self._account = account

        if self._account_css_class:
            self._account_identifier.remove_css_class(self._account_css_class)
            self._account_css_class = ""

        if not account:
            return

        self._account_css_class = app.css_config.get_dynamic_class(account)
        self._account_identifier.add_css_class(self._account_css_class)

    @GObject.Property(type=str)
    def jid(self) -> JID | None:  # type: ignore
        return self._jid

    @jid.setter
    def jid(self, jid: JID) -> None:
        self._jid = jid

    @GObject.Property(type=bool, default=False)
    def read(self) -> bool:  # type: ignore
        return self._read

    @read.setter
    def read(self, read: bool) -> None:
        self._read = read
        if self._read:
            self._activity_title_label.remove_css_class("bold")
        else:
            self._activity_title_label.add_css_class("bold")

    def bind(self, obj: ActivityListItem[E]) -> None:
        bind_spec = [
            ("account", self, "account"),
            ("jid", self, "jid"),
            ("account_visible", self._account_identifier, "visible"),
            ("activity_type_icon", self._activity_type_image, "icon_name"),
            ("avatar", self._avatar_image, "paintable"),
            ("title", self._activity_title_label, "label"),
            ("timestamp_string", self._timestamp_label, "label"),
            ("subject", self._activity_subject_label, "label"),
            ("read", self, "read"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self.__bindings.append(bind)

    def unbind(self) -> None:
        for bind in self.__bindings:
            bind.unbind()
        self.__bindings.clear()

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Grid.do_unroot(self)
        app.check_finalize(self)


class GajimUpdate(ActivityListItem[events.GajimUpdateAvailable]):
    @classmethod
    def from_event(cls, event: events.GajimUpdateAvailable) -> GajimUpdate:
        scale = app.window.get_scale_factor()
        texture = app.app.avatar_storage.get_gajim_circle_icon(AvatarSize.ROSTER, scale)

        return cls(
            context_id=event.context_id,
            account="",
            jid=None,
            account_visible=False,
            activity_type=0,
            activity_type_icon="lucide-info-symbolic",
            avatar=texture,
            title=_("Gajim Update"),
            timestamp=utc_now(),
            subject=_("Gajim %s is available") % event.version,
            read=False,
            event=event,
            unique=True,
        )


class GajimUpdatePermission(ActivityListItem[events.AllowGajimUpdateCheck]):
    @classmethod
    def from_event(cls, event: events.AllowGajimUpdateCheck) -> GajimUpdatePermission:
        scale = app.window.get_scale_factor()
        texture = app.app.avatar_storage.get_gajim_circle_icon(AvatarSize.ROSTER, scale)

        return cls(
            context_id=event.context_id,
            account="",
            jid=None,
            account_visible=False,
            activity_type=0,
            activity_type_icon="lucide-info-symbolic",
            avatar=texture,
            title=_("Gajim Update Check"),
            timestamp=utc_now(),
            subject=_("Search for Gajim updates periodically?"),
            read=False,
            event=event,
            unique=True,
        )


class GajimPluginUpdate(ActivityListItem[events.PluginUpdatesAvailable]):
    @classmethod
    def from_event(cls, event: events.PluginUpdatesAvailable) -> GajimPluginUpdate:
        scale = app.window.get_scale_factor()
        texture = app.app.avatar_storage.get_gajim_circle_icon(AvatarSize.ROSTER, scale)

        return cls(
            context_id=event.context_id,
            account="",
            jid=None,
            account_visible=False,
            activity_type=0,
            activity_type_icon="lucide-info-symbolic",
            avatar=texture,
            title=_("Plugin Update"),
            timestamp=utc_now(),
            subject=_("There are updates for Gajim’s plugins"),
            read=False,
            event=event,
            unique=True,
        )


class GajimPluginUpdateFinished(ActivityListItem[None]):
    @classmethod
    def from_event(cls) -> GajimPluginUpdateFinished:
        scale = app.window.get_scale_factor()
        texture = app.app.avatar_storage.get_gajim_circle_icon(AvatarSize.ROSTER, scale)

        return cls(
            context_id="",
            account="",
            jid=None,
            account_visible=False,
            activity_type=0,
            activity_type_icon="lucide-info-symbolic",
            avatar=texture,
            title=_("Plugins Updated Successfully"),
            timestamp=utc_now(),
            subject=_("Updates will be installed next time Gajim is started"),
            read=False,
            event=None,
            unique=True,
        )


class Subscribe(ActivityListItem[events.SubscribePresenceReceived]):
    @classmethod
    def from_event(cls, event: events.SubscribePresenceReceived) -> Subscribe:
        client = app.get_client(event.account)
        contact = client.get_module("Contacts").get_contact(event.jid)

        scale = app.window.get_scale_factor()
        assert not isinstance(contact, ResourceContact)
        texture = contact.get_avatar(AvatarSize.ROSTER, scale)

        nick = event.user_nick
        if not nick:
            assert isinstance(contact, BareContact)
            nick = contact.name

        subject = _("%s wants to add you.") % nick
        if event.status:
            subject = f"{subject} {event.status}"

        return cls(
            context_id=event.context_id,
            account=event.account,
            jid=event.jid,
            account_visible=app.app.multi_account_mode,
            activity_type=0,
            activity_type_icon="lucide-users-symbolic",
            avatar=texture,
            title=_("Contact Request"),
            timestamp=utc_now(),
            subject=subject,
            read=False,
            event=event,
        )

    def should_notify(self) -> bool:
        return True


class Unsubscribed(ActivityListItem[events.UnsubscribedPresenceReceived]):
    @classmethod
    def from_event(cls, event: events.UnsubscribedPresenceReceived) -> Unsubscribed:
        client = app.get_client(event.account)
        contact = client.get_module("Contacts").get_contact(event.jid)

        scale = app.window.get_scale_factor()
        assert not isinstance(contact, ResourceContact)
        texture = contact.get_avatar(AvatarSize.ROSTER, scale)

        subject = _("%s stopped sharing their status") % contact.name

        return cls(
            context_id=event.context_id,
            account=event.account,
            jid=event.jid,
            account_visible=app.app.multi_account_mode,
            activity_type=0,
            activity_type_icon="lucide-users-symbolic",
            avatar=texture,
            title=_("Contact Removed"),
            timestamp=utc_now(),
            subject=subject,
            read=False,
            event=event,
        )


class MucInvitation(ActivityListItem[events.MucInvitation]):
    @classmethod
    def from_event(cls, event: events.MucInvitation) -> MucInvitation:
        client = app.get_client(event.account)
        muc_contact = client.get_module("Contacts").get_contact(event.muc)
        assert isinstance(muc_contact, GroupchatContact)
        if muc_contact.muc_context == "private" and not event.muc.bare_match(
            event.from_
        ):
            contact = client.get_module("Contacts").get_contact(event.from_.bare)
            assert isinstance(contact, BareContact)
            subject = _("%(contact)s invited you to %(chat)s") % {
                "contact": contact.name,
                "chat": event.info.muc_name,
            }

        else:
            subject = _("You have been invited to %s") % event.info.muc_name

        if event.info.jid is not None:
            contact = client.get_module("Contacts").get_contact(event.info.jid)
        else:
            contact = client.get_module("Contacts").get_contact(event.from_.bare)

        scale = app.window.get_scale_factor()
        assert not isinstance(contact, ResourceContact)
        texture = contact.get_avatar(AvatarSize.ROSTER, scale)

        return cls(
            context_id=event.context_id,
            account=event.account,
            jid=event.from_,
            account_visible=app.app.multi_account_mode,
            activity_type=0,
            activity_type_icon="lucide-users-symbolic",
            avatar=texture,
            title=_("Invitation"),
            timestamp=utc_now(),
            subject=subject,
            read=False,
            event=event,
        )

    def should_notify(self) -> bool:
        return True


class MucInvitationDeclined(ActivityListItem[events.MucDecline]):
    @classmethod
    def from_event(cls, event: events.MucDecline) -> MucInvitationDeclined:
        client = app.get_client(event.account)
        contact = client.get_module("Contacts").get_contact(event.from_.bare)

        scale = app.window.get_scale_factor()
        assert not isinstance(contact, ResourceContact)
        texture = contact.get_avatar(AvatarSize.ROSTER, scale)

        muc_name = get_groupchat_name(client, event.muc)
        subject = _("%(contact)s declined your invitation to %(chat)s") % {
            "contact": contact.name,
            "chat": muc_name,
        }
        if event.reason is not None:
            subject = f"{subject} ({event.reason})"

        return cls(
            context_id=event.context_id,
            account=event.account,
            jid=event.from_,
            account_visible=app.app.multi_account_mode,
            activity_type=0,
            activity_type_icon="lucide-users-symbolic",
            avatar=texture,
            title=_("Invitation Declined"),
            timestamp=utc_now(),
            subject=subject,
            read=False,
            event=event,
        )


class OpenPGPEvent(ActivityListItem[events.OpenPGPEvent]):
    @classmethod
    def from_event(cls, event: events.OpenPGPEvent) -> OpenPGPEvent:
        scale = app.window.get_scale_factor()
        texture = app.app.avatar_storage.get_gajim_circle_icon(AvatarSize.ROSTER, scale)

        if event.type == "setup":
            title = _("OpenPGP Backup")
            subject = _("Complete the OpenPGP setup")
        else:
            title = _("OpenPGP Backup Error")
            subject = _("OpenPGP key backup failed")

        return cls(
            context_id=event.context_id,
            account=event.account,
            jid=None,
            account_visible=app.app.multi_account_mode,
            activity_type=0,
            activity_type_icon="lucide-info-symbolic",
            avatar=texture,
            title=title,
            timestamp=utc_now(),
            subject=subject,
            read=False,
            event=event,
            unique=True,
        )


class TimezoneChanged(ActivityListItem[events.TimezoneChanged]):
    @classmethod
    def from_event(cls, event: events.TimezoneChanged) -> TimezoneChanged:
        scale = app.window.get_scale_factor()
        texture = app.app.avatar_storage.get_gajim_circle_icon(AvatarSize.ROSTER, scale)
        return cls(
            context_id=event.context_id,
            account=event.account,
            jid=None,
            account_visible=app.app.multi_account_mode,
            activity_type=0,
            activity_type_icon="lucide-info-symbolic",
            avatar=texture,
            title=_("Timezone Update"),
            timestamp=utc_now(),
            subject=_("Update your timezone?"),
            read=False,
            event=event,
            unique=True,
        )


class Reaction(ActivityListItem[events.ReactionUpdated]):
    @classmethod
    def from_event(cls, event: events.ReactionUpdated) -> Reaction:
        client = app.get_client(event.account)
        scale = app.window.get_scale_factor()
        assert event.message is not None
        if event.message.type in (MessageType.GROUPCHAT, MessageType.PM):
            muc_jid = event.message.remote.jid
            if event.message.type == MessageType.PM:
                muc_jid.new_as_bare()

            groupchat_name = get_groupchat_name(client, muc_jid)

            assert event.occupant is not None
            texture = app.app.avatar_storage.get_occupant_texture(
                event.jid, event.occupant, AvatarSize.ROSTER, scale
            )
            nickname = event.occupant.nickname
            title = _("Reaction from %(nickname)s (%(chat)s)") % {
                "nickname": nickname,
                "chat": groupchat_name,
            }

        else:
            contact = client.get_module("Contacts").get_contact(event.jid)
            assert isinstance(contact, BareContact)
            texture = contact.get_avatar(AvatarSize.ROSTER, scale)
            nickname = contact.name
            title = _("Reaction from %s") % nickname

        assert event.message is not None
        assert event.emojis is not None

        # Add emoji presentation selector, otherwise depending on the font
        # emojis might be displayed in its text variant
        emojis = " ".join({f"{e}\ufe0f" for e in event.emojis})

        subject = _(
            "%(contact)s reacted with %(reaction)s to your message '%(message)s'"
        ) % {
            "contact": nickname,
            "reaction": emojis,
            "message": event.message.text or "",
        }

        read = app.window.is_chat_active(event.account, event.jid)

        return cls(
            context_id=event.context_id,
            account=event.account,
            jid=event.jid,
            account_visible=app.app.multi_account_mode,
            activity_type=0,
            activity_type_icon="lucide-reply-symbolic",
            avatar=texture,
            title=title,
            timestamp=utc_now(),
            subject=subject,
            read=read,
            event=event,
        )

    @staticmethod
    def can_create(event: events.ReactionUpdated) -> bool:
        return (
            event.message is not None
            # Message is in the database
            and event.message.direction == ChatDirection.OUTGOING
            # Message we sent out
            and event.direction == ChatDirection.INCOMING
            # A reaction we received from someone else
        )

    def activated(self) -> None:
        assert self.event.message is not None
        app.window.scroll_to_message(self.event.account, self.event.message)

    def should_notify(self) -> bool:
        assert self.event.message is not None
        if self.event.is_mam_message or app.window.is_chat_active(
            self.account, self.event.jid
        ):
            return False

        if self.event.message.type in (MessageType.GROUPCHAT, MessageType.PM):
            return app.settings.get_app_setting("gc_notify_on_reaction_default")
        return app.settings.get_app_setting("notify_on_reaction_default")


ActivityListItemT = (
    GajimUpdate
    | GajimUpdatePermission
    | GajimPluginUpdate
    | GajimPluginUpdateFinished
    | Subscribe
    | Unsubscribed
    | MucInvitation
    | MucInvitationDeclined
    | OpenPGPEvent
    | TimezoneChanged
    | Reaction
)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Literal
from typing import overload

import logging
from pathlib import Path

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common import passwords
from gajim.common.helpers import generate_qr_code
from gajim.plugins.plugins_i18n import _

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.filechoosers import FileChooserButton
from gajim.gtk.preference.widgets import CopyButton
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.openpgp.secret")


class OpenPGPSecret(Assistant):
    def __init__(self, account: str) -> None:
        Assistant.__init__(
            self,
            name="OpenPGPSecret",
            height=500,
            transient_for=app.app.get_active_window(),
            modal=True,
        )
        self._destroyed: bool = False

        self.account = account

        self.add_button("close", _("Close"))
        self.add_button(
            "unlock", _("Unlock"), complete=True, css_class="suggested-action"
        )

        self.add_pages(
            {
                "unlock": UnlockPage(account),
                "share": SharePage(account),
            }
        )

        self._connect(self, "button-clicked", self._on_assistant_button_clicked)

    @overload
    def get_page(self, name: Literal["unlock"]) -> UnlockPage: ...

    @overload
    def get_page(self, name: Literal["share"]) -> SharePage: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    def _on_assistant_button_clicked(
        self, _assistant: Assistant, button_name: str
    ) -> None:
        match button_name:
            case "unlock":
                if not self.get_page("unlock").check_unlock():
                    return
                self.get_page("share").generate()
                self.show_page("share")

            case "close":
                self.close()

            case _:
                raise ValueError(f"Unknown button: {button_name}")


@Gtk.Template(string=get_ui_string("openpgp/unlock.ui"))
class UnlockPage(AssistantPage):
    __gtype_name__ = "OpenPGPUnlock"

    _password_entry: Gtk.PasswordEntry = Gtk.Template.Child()
    _warning_box: Gtk.Box = Gtk.Template.Child()
    _warning_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, account: str) -> None:
        AssistantPage.__init__(self)
        self.title = _("Enter Password to Unlock")
        self.complete = True

        self._cur_password = passwords.get_password(account)

        self._connect(self._password_entry, "changed", self._on_changed)

    def _on_changed(self, _password_entry: Gtk.PasswordEntry) -> None:
        if self._cur_password is None:
            self.set_warning(_("No account password found, unlocking not possible"))
            self.set_complete(False)
            return

        self.set_warning(None)

    def check_unlock(self) -> bool:
        cur_password = self._password_entry.get_text()
        if cur_password != self._cur_password:
            self.set_warning(_("Password incorrect"))
            return False

        self.set_warning(None)
        return True

    def set_warning(self, text: str | None) -> None:
        self._warning_label.set_text(text or "")
        self._warning_box.set_visible(bool(text))

    def get_visible_buttons(self) -> list[str]:
        return ["close", "unlock"]

    def get_default_button(self) -> str:
        return "unlock"


@Gtk.Template(string=get_ui_string("openpgp/secret_share.ui"))
class SharePage(AssistantPage):
    __gtype_name__ = "OpenPGPSecretShare"

    _toast_overlay: Adw.ToastOverlay = Gtk.Template.Child()
    _box: Gtk.Box = Gtk.Template.Child()
    _key_label: Gtk.Label = Gtk.Template.Child()
    _copy_button: CopyButton = Gtk.Template.Child()
    _hint_label: Gtk.Label = Gtk.Template.Child()
    _qr_code_image: Gtk.Image = Gtk.Template.Child()

    def __init__(self, account: str) -> None:
        AssistantPage.__init__(self)
        self.set_valign(Gtk.Align.FILL)

        self.title = _("Share Secret Key")
        self.complete = True

        self._client = app.get_client(account)

        file_chooser_button = FileChooserButton(
            path=Path.home() / f"{self._client.account}-private.key",
            mode="save",
            label=_("Export…"),
            icon_name="lucide-file-down-symbolic",
        )
        self._box.append(file_chooser_button)
        self._connect(file_chooser_button, "path-picked", self._on_path_picked)
        self._connect(self._copy_button, "clicked", self._on_copy_clicked)

    def generate(self) -> None:
        share_uri = self._client.get_module("OpenPGP").get_secret_share_uri()
        backup_password = self._client.get_module("OpenPGP").get_backup_password()

        if share_uri is None or backup_password is None:
            self._key_label.set_text(_("No key found"))
            self._copy_button.set_sensitive(False)
            self._hint_label.set_visible(False)

            self._qr_code_image.set_visible(False)
            self._qr_code_image.set_from_paintable(None)

        else:
            self._key_label.set_text(backup_password)
            self._copy_button.set_sensitive(True)
            self._hint_label.set_visible(True)

            texture = generate_qr_code(share_uri)
            self._qr_code_image.set_visible(True)
            self._qr_code_image.set_from_paintable(texture)

    def _on_copy_clicked(self, _button: CopyButton) -> None:
        app.window.get_clipboard().set(self._key_label.get_text())
        self._toast_overlay.add_toast(
            Adw.Toast(title=_("Backup password copied"), timeout=1)
        )

    def _on_path_picked(self, button: Gtk.Button, paths: list[Path]) -> None:
        exported_key = self._client.get_module("OpenPGP").export_secret_key()
        if exported_key is None or not paths:
            return

        paths[0].write_text(exported_key)

        toast = Adw.Toast(
            title=_("Key has been saved"),
            timeout=5,
            button_label=_("Open Folder"),
            action_name="app.open-folder",
            action_target=GLib.Variant("s", str(paths[0])),
        )
        self._toast_overlay.add_toast(toast)

    def get_visible_buttons(self) -> list[str]:
        return ["close"]

    def get_default_button(self) -> str:
        return "close"



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging

import cairo
from gi.repository import Gdk
from gi.repository import Gsk
from gi.repository import Gtk

from gajim.common import app

from gajim.gtk.util.misc import convert_texture_to_surface

log = logging.getLogger("gajim.gtk.util.icons")


def icon_exists(name: str) -> bool:
    return get_icon_theme().has_icon(name)


def get_icon_theme() -> Gtk.IconTheme:
    display = Gdk.Display.get_default()
    assert display is not None
    return Gtk.IconTheme.get_for_display(display)


def find_texture_node(node: Any) -> Gsk.TextureNode | None:
    if isinstance(node, Gsk.TextureNode):
        return node

    if isinstance(node, Gsk.TransformNode):
        return find_texture_node(node.get_child())

    if isinstance(node, Gsk.ContainerNode):
        for i in range(node.get_n_children()):
            child = node.get_child(i)
            texture_node = find_texture_node(child)
            if texture_node is not None:
                return texture_node

    return None


def load_icon_surface(
    icon_name: str,
    size: int = 16,
    scale: int | None = None,
) -> cairo.ImageSurface | None:
    if scale is None:
        scale = app.window.get_scale_factor()

    if not scale:
        log.warning("Could not determine scale factor")
        scale = 1

    icon_theme = get_icon_theme()
    icon = icon_theme.lookup_icon(
        icon_name,
        None,
        size,
        scale,
        Gtk.TextDirection.NONE,
        0,  # type: ignore
    )

    snapshot = Gtk.Snapshot()
    icon.snapshot(snapshot, size, size)

    texture_node = find_texture_node(snapshot.to_node())
    if texture_node is None:
        log.error("Could not find texture node")
        return None

    texture = texture_node.get_texture()
    return convert_texture_to_surface(texture)


def get_tray_icon_name(name: str) -> str:
    prefix = "gajim"
    if app.is_flatpak():
        prefix = app.get_default_app_id()
    return f"{prefix}-status-{name}"


def get_account_tune_icon_name(account: str) -> str | None:
    client = app.get_client(account)
    tune = client.get_module("UserTune").get_current_tune()
    return None if tune is None else "lucide-music-symbolic"


def get_account_location_icon_name(account: str) -> str | None:
    client = app.get_client(account)
    location = client.get_module("UserLocation").get_current_location()
    return None if location is None else "lucide-map-pin-symbolic"



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from collections.abc import Callable

from gi.repository import Adw
from gi.repository import Gtk

from gajim.common import app

from gajim.gtk.util.classes import SignalManager


class ApplyButtonBox(Gtk.Box, SignalManager):
    def __init__(
        self, button_text: str, on_clicked: Callable[[Gtk.Button], Any]
    ) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        SignalManager.__init__(self)

        self._status_image = Gtk.Image(visible=False)
        self._spinner = Adw.Spinner(visible=False)
        self._button = Gtk.Button(label=button_text, sensitive=False)
        self._button.add_css_class("suggested-action")
        self._connect(self._button, "clicked", self._on_clicked)
        self._connect(self._button, "clicked", on_clicked)

        self.append(self._button)
        self.append(self._status_image)
        self.append(self._spinner)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _on_clicked(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        self._spinner.set_visible(True)

    def set_button_state(self, state: bool) -> None:
        if state:
            self._status_image.set_visible(False)
        self._button.set_sensitive(state)

    def set_success(self) -> None:
        self._spinner.set_visible(False)
        self._set_status_image("success")

    def set_error(self, tooltip_text: str):
        self._spinner.set_visible(False)
        self._set_status_image("error", tooltip_text)
        self._button.set_sensitive(True)

    def _set_status_image(self, state: str, tooltip_text: str = "") -> None:
        self._status_image.remove_css_class("success")
        self._status_image.remove_css_class("warning")

        icon_name = "lucide-check-symbolic"
        css_class = "success"

        if state == "error":
            icon_name = "lucide-circle-alert-symbolic"
            css_class = "warning"

        self._status_image.set_from_icon_name(icon_name)
        self._status_image.add_css_class(css_class)
        self._status_image.set_tooltip_text(tooltip_text)
        self._status_image.set_visible(True)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Final

import random
import secrets
import subprocess
from base64 import b64encode
from collections.abc import Generator
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path

import pysequoia as pys
from nbxmpp import Node
from nbxmpp import StanzaMalformed
from nbxmpp.client import Client as nbxmppClient
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.exceptions import StanzaDecrypted
from nbxmpp.modules.openpgp import create_message_stanza
from nbxmpp.modules.openpgp import create_signcrypt_node
from nbxmpp.modules.openpgp import parse_signcrypt
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.structs import DiscoItems
from nbxmpp.structs import EncryptionData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import PGPKeyMetadata
from nbxmpp.structs import PGPPublicKey
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import Task

import gajim.common.storage.openpgp.models as mod
from gajim.common import app
from gajim.common import configpaths
from gajim.common import ged
from gajim.common import types
from gajim.common.const import EncryptionInfoMsg
from gajim.common.const import Trust
from gajim.common.events import EncryptionInfo
from gajim.common.events import MucAdded
from gajim.common.events import OpenPGPEvent
from gajim.common.events import SignedIn
from gajim.common.helpers import timeout_add_seconds_once
from gajim.common.i18n import _
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.message_util import get_chat_type_and_direction
from gajim.common.modules.message_util import get_message_timestamp
from gajim.common.modules.util import as_task
from gajim.common.modules.util import CryptoModule
from gajim.common.modules.util import event_node
from gajim.common.modules.util import PublicKeyData
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.structs import OutgoingMessage
from gajim.common.util.decorators import cache_with_ttl
from gajim.common.util.decorators import event_filter
from gajim.common.util.text import format_fingerprint

NOT_ENCRYPTED_TAGS = [
    ("no-store", Namespace.HINTS),
    ("store", Namespace.HINTS),
    ("no-copy", Namespace.HINTS),
    ("no-permanent-store", Namespace.HINTS),
    ("origin-id", Namespace.SID),
    ("thread", ""),
]
BACKUP_PASSWORD_ALPHABET: Final = "123456789ABCDEFGHIJKLMNPQRSTUVWXYZ"  # noqa: S105


class NoSecretKeyFound(Exception):
    def __init__(self) -> None:
        super().__init__(_("No secret key found"))


class NoSecretKeyImportError(Exception):
    def __init__(self) -> None:
        super().__init__(_("Imported key is not a secret key"))


class SecretKeyExpirationImportError(Exception):
    def __init__(self) -> None:
        super().__init__(_("Imported key has expiration date"))


class MultipleSecretKeysImportError(Exception):
    def __init__(self, secret_certs: list[pys.Cert]) -> None:
        self._secret_certs = secret_certs

    def get_certs(self) -> list[pys.Cert]:
        return self._secret_certs


class BackupDecryptionError(Exception):
    def __init__(self) -> None:
        super().__init__(_("Unable to decrypt the backup with this password"))


class EncryptionTestImportError(Exception):
    def __init__(self) -> None:
        super().__init__(_("Unable to use this key for encrypting messages"))


class DecryptionFailed(Exception):
    pass


def prepare_stanza(stanza: Node, payload: list[Node | str]) -> None:
    delete_nodes(stanza, "openpgp", Namespace.OPENPGP)
    delete_nodes(stanza, "body")

    nodes: list[Node] = []
    for node in payload:
        if isinstance(node, str):
            continue
        name, namespace = node.getName(), node.getNamespace()
        delete_nodes(stanza, name, namespace)
        nodes.append(node)

    for node in nodes:
        stanza.addChild(node=node)


def delete_nodes(stanza: Node, name: str, namespace: str | None = None) -> None:
    attrs = None
    if namespace is not None:
        attrs = {"xmlns": Namespace.OPENPGP}
    nodes = stanza.getTags(name, attrs)
    for node in nodes:
        stanza.delChild(node)


def find_remote_key(
    public_rows: Sequence[mod.Public], fingerprint: str
) -> mod.Public | None:
    for public in public_rows:
        if public.fingerprint == fingerprint:
            return public


def find_cert_with_fpr(
    certs: list[pys.Cert], fingerprint: str | None
) -> pys.Cert | None:
    if fingerprint is None:
        return certs[0]

    for cert in certs:
        if cert.fingerprint == fingerprint:
            return cert
    return None


@dataclass
class OpenPGPPublicKeyData(PublicKeyData):
    @classmethod
    def from_public(cls, public: mod.Public) -> OpenPGPPublicKeyData:
        return cls(
            active=public.active,
            address=public.remote_jid,
            label=public.label,
            last_seen=public.last_seen,
            fingerprint=public.fingerprint,
            trust=Trust(public.trust),
        )

    def pretty_fingerprint(self, wrap: bool = True) -> str:
        return format_fingerprint(self.fingerprint, "OpenPGP", wrap=wrap)


class OpenPGP(BaseModule, CryptoModule):
    _nbxmpp_extends = "OpenPGP"
    _nbxmpp_methods = [
        "set_keylist",
        "request_keylist",
        "set_public_key",
        "request_public_key",
        "set_secret_key",
        "request_secret_key",
    ]

    def __init__(self, client: types.Client):
        BaseModule.__init__(self, client)

        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self.decrypt_message,
                ns=Namespace.OPENPGP,
                priority=9,
            ),
        ]

        self._register_pubsub_handler(self._keylist_notification_received)
        self.register_events(
            [
                ("signed-in", ged.CORE, self._on_signed_in),
                ("muc-added", ged.PREGUI, self._on_muc_added),
            ]
        )

        self._own_jid = self._get_own_bare_jid()
        self._secret_cert = None
        self._backup_password = None

        self._load_secret_keys()
        self._migrate_secret_keys()

        timeout_add_seconds_once(random.randint(30, 60), self.check_secret_key_backup)

        app.settings.connect_signal(
            "openpgp_backup_secret_key", self._on_backup_settng_changed, self._account
        )

    def _on_backup_settng_changed(self, enabled: bool, *args: Any) -> None:
        if enabled:
            self.check_secret_key_backup()

    def _load_secret_keys(self) -> None:
        if row := app.storage.openpgp.get_secret_key(self._account):
            secret, self._secret_key_date = row
            self._secret_cert = pys.Cert.from_bytes(secret.key)
            self._backup_password = secret.backup_password

            self._log.info(
                "Found secret key: %s, %s",
                self._secret_cert.fingerprint.upper(),
                self._secret_key_date,
            )

    def _migrate_secret_keys(self) -> None:
        if self._secret_cert is not None:
            return

        plugin_data_path = (
            Path(configpaths.get("MY_DATA")) / "openpgp" / str(self._own_jid)
        )
        if not plugin_data_path.exists():
            return

        self._log.info("Try to migrate secret keys from openpgp plugin")

        try:
            result = subprocess.check_output(  # noqa: S603
                [  # noqa: S607
                    "gpg",
                    "--homedir",
                    plugin_data_path.resolve(),
                    "--armor",
                    "--export-secret-key",
                ]
            )
        except Exception as error:
            self._log.warning("Unable to migrate secret keys: %s", error)
            return

        try:
            certs = pys.Cert.split_bytes(result)
        except Exception as error:
            self._log.warning("Failed to convert keys to certs: %s", error)
            return

        for cert in certs:
            user_ids = set(map(str, cert.user_ids))
            if self._own_jid.to_iri() not in user_ids:
                continue

            if not cert.has_secret_keys:
                continue

            assert cert.secrets is not None
            backup_password = self.generate_backup_password()
            app.storage.openpgp.store_secret_key(
                self._account, cert, backup_password=backup_password
            )
            self._log.info(
                "Successfully migrated secret key: %s", cert.fingerprint.upper()
            )
            self._load_secret_keys()
            break

    @event_filter(["account"])
    def _on_signed_in(self, _event: SignedIn) -> None:
        if self.secret_key_exists():
            self.request_keylist()
            self.set_public_key()

    def secret_key_exists(self) -> bool:
        return self._secret_cert is not None

    def generate_key(self) -> None:
        assert self._secret_cert is None
        cert = pys.Cert.generate(
            self._own_jid.to_iri(), profile=pys.Profile.RFC4880, validity_seconds=None
        )
        assert cert.secrets is not None

        backup_password = self.generate_backup_password()
        app.storage.openpgp.store_secret_key(
            self._account, cert, backup_password=backup_password
        )
        self._log.info("Generated key %s", cert.fingerprint.upper())
        self._load_secret_keys()
        self.set_public_key()
        self.request_keylist()
        self.check_secret_key_backup()

    def get_backup_password(self) -> str | None:
        return self._backup_password

    def test_backup_password(self, data: bytes, password: str) -> None:
        try:
            decrypted = pys.decrypt(data, passwords=[password])
        except Exception:
            raise BackupDecryptionError

        if decrypted.bytes is None:
            raise BackupDecryptionError

        pys.Cert.split_bytes(decrypted.bytes)

        self._backup_password = password
        app.storage.openpgp.store_secret_key_backup_password(self._account, password)
        self.check_secret_key_backup()

    def import_key(
        self,
        data: str | bytes,
        backup_password: str | None = None,
        fingerprint: str | None = None,
    ) -> None:
        if self._secret_cert is not None:
            raise ValueError("A secret key already exists")

        if isinstance(data, str):
            data = data.encode()

        if backup_password is not None:
            decrypted = pys.decrypt(data, passwords=[backup_password])
            if decrypted.bytes is None:
                raise NoSecretKeyImportError

            data = decrypted.bytes

        secret_certs = pys.Cert.split_bytes(data)
        self._log.info("Certs in import data: %s", secret_certs)

        if len(secret_certs) > 1 and not fingerprint:
            raise MultipleSecretKeysImportError(secret_certs)

        cert = find_cert_with_fpr(secret_certs, fingerprint)
        if cert is None:
            raise NoSecretKeyImportError

        if not cert.has_secret_keys:
            self._log.warning(
                "Found public key in secret key backup: %s", cert.fingerprint
            )
            raise NoSecretKeyImportError

        if cert.expiration is not None:
            raise SecretKeyExpirationImportError

        assert cert.secrets is not None

        try:
            pys.encrypt(b"", recipients=[cert], signer=cert.secrets.signer())
        except Exception:
            raise EncryptionTestImportError

        user_ids = set(map(str, cert.user_ids))
        if self._own_jid.to_iri() not in user_ids:
            cert = cert.add_user_id(
                value=self._own_jid.to_iri(), certifier=cert.secrets.certifier()
            )

        assert cert.secrets is not None
        if not backup_password:
            backup_password = self.generate_backup_password()

        app.storage.openpgp.store_secret_key(self._account, cert, backup_password)
        self._load_secret_keys()
        self.set_public_key()
        self.request_keylist()
        self.check_secret_key_backup()

    def set_public_key(self) -> None:
        self._log.info("Publish public key")

        assert self._secret_cert is not None
        # Make sure we dont accidentally publish the secret key
        public_cert = pys.Cert.from_bytes(bytes(self._secret_cert))
        assert not public_cert.has_secret_keys

        self._nbxmpp("OpenPGP").set_public_key(
            bytes(public_cert),
            self._secret_cert.fingerprint.upper(),
            self._secret_key_date.timestamp(),
        )

    def request_public_key(self, jid: JID, fingerprint: str) -> None:
        self._log.info("Request public key %s - %s", fingerprint, jid)
        self._nbxmpp("OpenPGP").request_public_key(
            jid, fingerprint, callback=self._public_key_received, user_data=fingerprint
        )

    def _public_key_received(self, task: Task) -> None:
        fingerprint = task.get_user_data()
        try:
            result = cast(PGPPublicKey | None, task.finish())
        except (StanzaError, MalformedStanzaError) as error:
            self._log.error("Public Key %s not found: %s", fingerprint, error)
            return

        if result is None:
            self._log.error("Public Key Node %s is empty", fingerprint)
            return

        self._log.info("Received public key: %s", fingerprint)

        try:
            cert = pys.Cert.from_bytes(result.key)
        except Exception:
            self._log.exception("Unable to parse public key")
            return

        user_ids = set(map(str, cert.user_ids))
        if result.jid.to_iri() not in user_ids:
            self._log.warning(
                "Ignore public key because of invalid user id: %s, user ids: %s",
                cert.fingerprint.upper(),
                user_ids,
            )
            return

        trust = Trust.UNDECIDED
        if app.settings.get_account_setting(self._account, "openpgp_blind_trust"):
            trust = Trust.BLIND

        app.storage.openpgp.store_public_key(self._account, result.jid, cert, trust)

    def set_keylist(self, keylist: list[PGPKeyMetadata] | None = None) -> None:
        if keylist is None:
            assert self._secret_cert is not None
            keylist = [
                PGPKeyMetadata(
                    self._own_jid,
                    self._secret_cert.fingerprint.upper(),
                    self._secret_key_date.timestamp(),
                )
            ]

        self._log.info("Publish keylist")
        self._nbxmpp("OpenPGP").set_keylist(keylist)

    @event_node(Namespace.OPENPGP_PK)
    def _keylist_notification_received(
        self, _client: nbxmppClient, _stanza: Node, properties: MessageProperties
    ) -> None:
        assert properties.pubsub_event is not None
        assert properties.jid is not None

        keylist: list[PGPKeyMetadata] = []
        if properties.pubsub_event.data:
            keylist = cast(list[PGPKeyMetadata], properties.pubsub_event.data)

        self._process_keylist(keylist, properties.jid)

    @cache_with_ttl(ttl=7200)
    def _request_keylist_ttl(self, remote_jid: JID) -> None:
        self.request_keylist(remote_jid)

    def request_keylist(self, jid: JID | None = None) -> None:
        if jid is None:
            jid = self._own_jid

        self._log.info("Fetch keylist %s", jid)

        self._nbxmpp("OpenPGP").request_keylist(
            jid, callback=self._keylist_received, user_data=jid
        )

    def _keylist_received(self, task: Task) -> None:
        jid = cast(JID, task.get_user_data())
        try:
            keylist = cast(list[PGPKeyMetadata] | None, task.finish())
        except (StanzaError, MalformedStanzaError) as error:
            self._log.error("Keylist query failed: %s", error)
            if self._own_jid.bare_match(jid) and self._secret_cert is not None:
                self.set_keylist()
            return

        self._log.info("Keylist received from %s", jid)
        self._process_keylist(keylist, jid)

    def _validate_active_fingerprints(
        self, keylist: list[PGPKeyMetadata]
    ) -> tuple[set[str], bool]:
        own_fpr_found = False
        active_fprs: set[str] = set()
        for key in keylist:
            if not key.fingerprint.isupper():
                self._log.warning("lower-cased fingerprint: %s", key.fingerprint)
                continue

            if (
                self._secret_cert is not None
                and self._secret_cert.fingerprint.upper() == key.fingerprint
            ):
                # Remove own fingerprint from the list
                own_fpr_found = True
                continue

            active_fprs.add(key.fingerprint)

        return active_fprs, own_fpr_found

    def _process_keylist(
        self, keylist: list[PGPKeyMetadata] | None, from_jid: JID
    ) -> None:

        self._log.info("Received keylist from %s", from_jid)
        own_keylist = self._own_jid.bare_match(from_jid)

        if own_keylist and not self.secret_key_exists():
            self._log.info("Ignore own keylist because no secret key was found")
            return

        if not keylist:
            self._log.info("Keylist is empty")
            if own_keylist:
                self.set_keylist()
            app.storage.openpgp.update_public_keys(
                self._account, from_jid, active=False
            )
            return

        active_fprs, own_fpr_found = self._validate_active_fingerprints(keylist)
        self._log.info("Active keys: %s, own key found: %s", active_fprs, own_fpr_found)

        known_fprs = app.storage.openpgp.get_known_fingerprints(
            self._account, [from_jid]
        )

        unknown_fprs = active_fprs - known_fprs
        inactive_fprs = known_fprs - active_fprs

        if active_fprs:
            app.storage.openpgp.update_public_keys(
                self._account, from_jid, fingerprints=active_fprs, active=True
            )

        if inactive_fprs:
            app.storage.openpgp.update_public_keys(
                self._account, from_jid, fingerprints=inactive_fprs, active=False
            )

        for fingerprint in unknown_fprs:
            self.request_public_key(from_jid, fingerprint)

        if own_keylist and not own_fpr_found:
            assert self._secret_cert is not None
            self._log.info("Own public key not published")
            keylist.append(
                PGPKeyMetadata(
                    self._own_jid,
                    self._secret_cert.fingerprint.upper(),
                    self._secret_key_date.timestamp(),
                )
            )
            self.set_keylist(keylist)

    def _process_muc_message(self, properties: MessageProperties) -> str | None:
        assert properties.jid is not None
        resource = properties.jid.resource
        contact = self._client.get_module("Contacts").get_contact(properties.jid)
        assert isinstance(contact, GroupchatParticipant)
        if contact.real_jid is not None:
            return contact.real_jid.bare

        self._log.error("Unable to find jid for group chat message from %s", resource)
        return None

    def _get_signer_from_message(
        self, remote_jid: JID, properties: MessageProperties
    ) -> tuple[JID | None, ChatDirection | None]:
        muc_data = None
        if properties.type.is_groupchat:
            muc_data = self._client.get_module("MUC").get_muc_data(remote_jid)
            if muc_data is None:
                self._log.warning("Groupchat message from unknown MUC: %s", remote_jid)
                return None, None

        _, direction = get_chat_type_and_direction(muc_data, self._own_jid, properties)
        if direction == ChatDirection.OUTGOING:
            return self._own_jid, direction

        if not properties.from_muc:
            return properties.remote_jid, direction

        if properties.mam:
            self._log.info("Message received, archive: %s", properties.mam.archive)
            if properties.muc_user is None or properties.muc_user.jid is None:
                self._log.warning(
                    "Received MAM message which can not be mapped to a real jid"
                )
                return None, None
            return properties.muc_user.jid.new_as_bare(), direction

        else:
            assert properties.jid is not None
            contact = self._client.get_module("Contacts").get_contact(properties.jid)
            if isinstance(contact, GroupchatParticipant):
                if contact.real_jid is not None:
                    return contact.real_jid.new_as_bare(), direction

        return None, None

    def decrypt_message(
        self, _client: nbxmppClient, stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_openpgp:
            return

        assert properties.openpgp is not None

        if self._secret_cert is None:
            self._log.warning(
                "Received encrypted message but no secret key is available"
            )
            return

        self._log.info("Received encrypted message from %s", properties.remote_jid)

        remote_jid = properties.remote_jid
        assert remote_jid is not None
        assert self._secret_cert.secrets is not None
        assert self._secret_cert.has_secret_keys

        signer_jid, direction = self._get_signer_from_message(remote_jid, properties)
        if signer_jid is None:
            self._log.warning("Unable to determine remote jid for message")
            return

        remote_public_keys = app.storage.openpgp.get_public_keys(
            self._account,
            [signer_jid],
            only_active=False,
        )
        if self._own_jid == signer_jid:
            own_public_key = app.storage.openpgp.get_public_key_from_secret(
                self._account, self._own_jid, self._secret_cert
            )
            remote_public_keys.append(own_public_key)

        def _store(_key_id: list[str]) -> list[pys.Cert]:
            return [public.key for public in remote_public_keys]

        try:
            decrypted = pys.decrypt(
                bytes=properties.openpgp,
                decryptor=self._secret_cert.secrets.decryptor(),
                store=_store,
            )
        except Exception as error:
            self._log.warning(
                "Failed to decrypt with signature verification: %s", error
            )
            try:
                decrypted = pys.decrypt(
                    bytes=properties.openpgp,
                    decryptor=self._secret_cert.secrets.decryptor(),
                )
            except Exception as error:
                self._log.warning(
                    "Failed to decrypt without signature verification: %s", error
                )
                return

        assert decrypted.bytes is not None
        payload = decrypted.bytes.decode()

        signcrypt = Node(node=payload)

        try:
            payload, recipients, _timestamp = parse_signcrypt(signcrypt)
        except StanzaMalformed as error:
            self._log.warning("Malformed payload: %s", error)
            self._log.warning(payload)
            return

        to_address = self._own_jid
        if properties.type.is_groupchat or direction == ChatDirection.OUTGOING:
            to_address = remote_jid

        if to_address not in recipients:
            self._log.warning("to attr not valid")
            self._log.warning(signcrypt)
            return

        prepare_stanza(stanza, payload)

        trust = Trust.UNTRUSTED
        fingerprint = "Unknown"

        if decrypted.valid_sigs:
            fingerprint = decrypted.valid_sigs[0].certificate.upper()
            remote_key = find_remote_key(remote_public_keys, fingerprint)
            if remote_key is None:
                self._log.warning("Unable to find remote key: %s", fingerprint)
            else:
                app.storage.openpgp.set_last_seen(
                    self._account,
                    remote_key.remote_jid,
                    fingerprint,
                    get_message_timestamp(properties),
                )
                trust = remote_key.trust

        properties.encrypted = EncryptionData(
            protocol="OpenPGP", key=fingerprint, trust=trust
        )

        raise StanzaDecrypted

    def check_send_preconditions(self, contact: types.ChatContactT) -> bool:
        match contact:
            case GroupchatContact():
                if not self._is_compatible_muc(contact.jid):
                    app.ged.raise_event(
                        EncryptionInfo(
                            account=contact.account,
                            jid=contact.jid,
                            type=EncryptionInfoMsg.BAD_MUC_CONFIG,
                            message=EncryptionInfoMsg.BAD_MUC_CONFIG.value.format(
                                encryption="OpenPGP"
                            ),
                        )
                    )
                    return False

                return True

            case BareContact():
                if contact.is_self:
                    return self.secret_key_exists()

                active, undecided = self._remote_has_active_keys(contact.jid)
                if not active:
                    self.request_keylist(contact.jid)
                    app.ged.raise_event(
                        EncryptionInfo(
                            account=contact.account,
                            jid=contact.jid,
                            type=EncryptionInfoMsg.QUERY_DEVICES,
                            message=EncryptionInfoMsg.QUERY_DEVICES.value,
                        )
                    )
                    return False

                if undecided:
                    app.ged.raise_event(
                        EncryptionInfo(
                            account=contact.account,
                            jid=contact.jid,
                            type=EncryptionInfoMsg.UNDECIDED_FINGERPRINTS,
                            message=EncryptionInfoMsg.UNDECIDED_FINGERPRINTS.value,
                        )
                    )
                    return False
                return True

            case _:
                raise ValueError(f"Unhandled contact type: {contact!r}")

    @event_filter(["account"])
    def _on_muc_added(self, event: MucAdded) -> None:
        client = app.get_client(event.account)
        contact = client.get_module("Contacts").get_contact(event.jid)
        if not isinstance(contact, GroupchatContact):
            self._log.warning("%s is not a groupchat contact", contact)
            return

        # Event is triggert on every join, avoid multiple connects
        contact.disconnect_all_from_obj(self)
        contact.connect("room-joined", self._on_muc_event)
        contact.connect("user-affiliation-changed", self._on_muc_event)
        contact.connect("room-affiliation-changed", self._on_muc_event)
        contact.connect("room-affiliations-complete", self._on_muc_event)

    def _on_muc_event(
        self, contact: GroupchatContact, _signal_name: str, *args: Any
    ) -> None:

        if not self._is_compatible_muc(contact.jid):
            return

        self._check_groupchat_members_keys(contact.jid)

    def _get_groupchat_members(self, remote_jid: JID) -> set[JID]:
        affiliations = self._client.get_module("MUC").get_affiliations(
            remote_jid, include_outcast=False
        )
        members: set[JID] = set()
        members.update(*affiliations.values())
        return members

    def _check_groupchat_members_keys(self, remote_jid: JID) -> None:
        for jid in self._get_groupchat_members(remote_jid):
            if not self._is_contact_in_roster(jid):
                self._log.info("%s not in roster, query keylist...", jid)
                self._request_keylist_ttl(jid)

    def _is_contact_in_roster(self, remote_jid: JID) -> bool:
        if remote_jid == self._own_jid:
            return True

        roster_item = self._client.get_module("Roster").get_item(remote_jid)
        if roster_item is None:
            return False

        contact = self._client.get_module("Contacts").get_contact(remote_jid)
        assert isinstance(contact, BareContact)
        return contact.subscription == "both"

    def _is_compatible_muc(self, remote_jid: JID) -> bool:
        disco_info = app.storage.cache.get_last_disco_info(remote_jid)
        assert disco_info is not None
        return disco_info.muc_is_members_only and disco_info.muc_is_nonanonymous

    def _remote_has_active_keys(self, remote: JID) -> tuple[bool, bool]:
        public_keys = app.storage.openpgp.get_public_keys(
            self._account,
            [remote],
            trust=[Trust.VERIFIED, Trust.BLIND, Trust.UNDECIDED],
        )
        for key in public_keys:
            if key.trust == Trust.UNDECIDED:
                return True, True

        return bool(public_keys), False

    def encrypt_message(self, message: OutgoingMessage) -> None:
        if self._secret_cert is None:
            raise NoSecretKeyFound

        remote_jid = message.contact.jid

        jids = [self._own_jid]
        if message.is_groupchat:
            jids += self._get_groupchat_members(remote_jid)
        else:
            jids += [remote_jid]

        recipients = app.storage.openpgp.get_public_keys(
            self._account,
            jids,
            trust=[Trust.VERIFIED, Trust.BLIND],
        )

        recipient_certs = [public.key for public in recipients]
        recipient_certs.append(self._secret_cert)

        self._log.info(
            "Encrypt to %s:\n%s",
            remote_jid,
            "\n".join([r.fingerprint.upper() for r in recipient_certs]),
        )

        payload = create_signcrypt_node(
            message.get_stanza(), [remote_jid], NOT_ENCRYPTED_TAGS
        )
        payload = str(payload).encode("utf8")

        assert self._secret_cert.secrets is not None

        encrypted_payload = pys.encrypt(
            signer=self._secret_cert.secrets.signer(),
            recipients=recipient_certs,
            bytes=payload,
            armor=False,
        )

        assert encrypted_payload is not None

        create_message_stanza(
            message.get_stanza(), encrypted_payload, bool(message.get_text())
        )

        message.set_encryption(
            EncryptionData(
                protocol="OpenPGP",
                key=self._secret_cert.fingerprint.upper(),
                trust=Trust.VERIFIED,
            )
        )

    def generate_backup_password(self) -> str:
        return "-".join(
            "".join(secrets.choice(BACKUP_PASSWORD_ALPHABET) for _ in range(4))
            for _ in range(6)
        )

    def get_secret_share_uri(self) -> str | None:
        if self._secret_cert is None or self._backup_password is None:
            return None

        assert self._secret_cert.secrets is not None
        cert_bytes = bytes(self._secret_cert.secrets)
        encrypted_key = pys.encrypt(
            passwords=[self._backup_password], bytes=cert_bytes, armor=False
        )
        query = (
            "openpgp",
            [
                ("secret", b64encode(encrypted_key).decode()),
                ("password", self._backup_password),
            ],
        )
        return self._own_jid.to_iri(query)

    def export_secret_key(self) -> str | None:
        if self._secret_cert is None:
            return None
        return str(self._secret_cert.secrets)

    @as_task
    def backup_secret_key(
        self, cert_bytes: bytes | None = None
    ) -> Generator[Any, None]:
        _task = yield  # noqa: F841

        if self._secret_cert is None:
            self._log.info("No secret key found, backup not possible")
            return

        self._log.info("Backup secret key")
        assert self._backup_password is not None

        if cert_bytes is None:
            assert self._secret_cert.secrets is not None
            cert_bytes = bytes(self._secret_cert.secrets)

        encrypted_payload = pys.encrypt(
            passwords=[self._backup_password], bytes=cert_bytes, armor=False
        )

        result = yield self.set_secret_key(encrypted_payload)
        raise_if_error(result)

    def remove_backup_from_server(self) -> None:
        if not app.account_is_available(self._account):
            return

        self._log.info("Remove backup from server")
        self._client.get_module("PubSub").delete(Namespace.OPENPGP_SK, self._own_jid)

    def check_secret_key_backup(self) -> None:
        if not app.account_is_available(self._account):
            return

        if not app.settings.get_account_setting(
            self._account, "openpgp_backup_secret_key"
        ):
            return

        self._log.info("Check secret key backup")
        self.request_secret_key(callback=self._secret_key_received)

    def _secret_key_received(self, task: Task) -> None:
        try:
            encrypted_bytes = cast(bytes | None, task.finish())
        except StanzaError as error:
            if error.condition == "item-not-found":
                self.backup_secret_key()
            else:
                self._log.error("Error on secret key request: %s", error)
                if self.secret_key_exists():
                    app.ged.raise_event(
                        OpenPGPEvent(
                            account=self._account,
                            type="unknown-error",
                            error=str(error),
                        )
                    )
            return

        except Exception as error:
            self._log.error("Error on secret key request: %s", error)
            if self.secret_key_exists():
                app.ged.raise_event(
                    OpenPGPEvent(
                        account=self._account, type="unknown-error", error=str(error)
                    )
                )
            return

        if not encrypted_bytes:
            self._log.info("No secret key backup found")
            self.backup_secret_key()
            return

        if not self.secret_key_exists():
            app.ged.raise_event(OpenPGPEvent(self._account, type="setup"))
            return

        if self._backup_password is None:
            self._log.info("Backup password not available, unable to decrypt backup")
            return

        try:
            decrypted = pys.decrypt(
                passwords=[self._backup_password], bytes=encrypted_bytes
            )
        except Exception as error:
            self._log.warning("Unable to decrypt secret key backup: %s", error)
            app.ged.raise_event(OpenPGPEvent(self._account, type="decryption-error"))
            return

        if decrypted.bytes is None:
            self._log.warning("No bytes after decryption")
            self.backup_secret_key()
            return

        secret_certs = pys.Cert.split_bytes(decrypted.bytes)
        self._log.info("Found certs in server backup: %s", secret_certs)

        assert self._secret_cert is not None
        assert self._secret_cert.secrets is not None

        for cert in secret_certs:
            if not cert.has_secret_keys:
                self._log.warning(
                    "Found public key in secret key backup: %s", cert.fingerprint
                )
                continue

            if cert.fingerprint == self._secret_cert.fingerprint:
                self._log.info("Found our secret key server backup")
                return

        self._log.info("Secret key not found in server backup")
        our_cert_bytes = bytes(self._secret_cert.secrets)
        self.backup_secret_key(decrypted.bytes + our_cert_bytes)

    def get_our_public_key(self) -> OpenPGPPublicKeyData | None:
        if self._secret_cert is None:
            return None

        return OpenPGPPublicKeyData(
            active=True,
            address=self._own_jid,
            label=None,
            last_seen=None,
            fingerprint=self._secret_cert.fingerprint.upper(),
            trust=Trust.VERIFIED,
        )

    def get_public_keys(self, jid: JID, *, is_groupchat: bool) -> list[PublicKeyData]:

        trust = [Trust.UNDECIDED] if is_groupchat else None
        public_key_data = app.storage.openpgp.get_public_keys(
            self._account, [jid], trust=trust, only_active=False
        )
        return [OpenPGPPublicKeyData.from_public(public) for public in public_key_data]

    def compose_trust_uri(self, jid: JID) -> str:
        return ""

    def remove_public_key(self, public_key_data: PublicKeyData) -> None:
        app.storage.openpgp.remove_public_key(
            self._account, public_key_data.address, public_key_data.fingerprint
        )

    def set_public_key_trust(
        self, public_key_data: PublicKeyData, trust: Trust
    ) -> None:
        app.storage.openpgp.update_public_keys(
            self._account,
            public_key_data.address,
            [public_key_data.fingerprint],
            trust=trust,
        )

    def clear_keylist(self) -> None:
        self._client.get_module("Discovery").disco_items(
            self._own_jid, callback=self._on_disco_items_result
        )

    def _on_disco_items_result(self, task: Task) -> None:
        try:
            result = cast(DiscoItems, task.finish())
        except Exception as error:
            self._log.warning("Disco items failed: %s", error)

        else:
            for item in result.items:
                if item.node is None or not item.node.startswith(
                    f"{Namespace.OPENPGP_PK}:"
                ):
                    continue
                self._log.info("Remove pubsub node: %s", item.node)
                self._client.get_module("PubSub").delete(item.node, self._own_jid)

        if self.secret_key_exists():
            self.set_public_key()
            self.set_keylist()



# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Any

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import MDSData
from nbxmpp.structs import MessageProperties

from gajim.common import app
from gajim.common import types
from gajim.common.events import ReadStateSync
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import event_node


class MDS(BaseModule):
    _nbxmpp_extends = "MDS"
    _nbxmpp_methods = ["set_mds"]

    def __init__(self, client: types.Client):
        BaseModule.__init__(self, client)
        self._register_pubsub_handler(self._mds_received)

    @event_node(Namespace.MDS)
    def _mds_received(
        self, _client: types.Client, _stanza: Any, properties: MessageProperties
    ) -> None:
        if not properties.pubsub_event:
            return

        data = properties.pubsub_event.data

        if data is None:
            return

        assert isinstance(data, MDSData)

        if not data.jid or not data.stanza_id:
            self._log.warning("Received invalid MDS event")
            return

        self._log.info("Read state sync (MDS): %s - %s", data.jid, data.stanza_id)
        app.ged.raise_event(
            ReadStateSync(account=self._account, jid=data.jid, marker_id=data.stanza_id)
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import io
import threading

from nbxmpp.util import utf8_decode
from PIL import Image

from gajim.common.multiprocess.http import http_request
from gajim.common.open_graph_parser import OpenGraphData
from gajim.common.open_graph_parser import OpenGraphParser
from gajim.common.open_graph_parser import OpenGraphThumbnail

THUMBNAIL_SIZE = 256


def generate_url_preview(
    url: str,
    event: threading.Event,
    proxy: str | None,
    http2: bool,
) -> OpenGraphData | None:
    result = http_request(
        event=event,
        ft_id="preview",
        method="GET",
        url=url,
        timeout=5,
        max_download_size=1024 * 150,
        proxy=proxy,
        http2=http2,
    )

    html_content, _ = utf8_decode(result.content)

    open_graph_parser = OpenGraphParser()
    og_data = open_graph_parser.parse(html_content)
    if og_data is None or not og_data.image:
        return og_data

    try:
        image_result = http_request(
            event=event,
            ft_id="preview",
            method="GET",
            url=og_data.image,
            timeout=5,
            proxy=proxy,
            http2=http2,
        )

        og_data.thumbnail = _make_thumbnail(image_result.content)
    except Exception:
        pass

    return og_data


def _make_thumbnail(content: bytes) -> OpenGraphThumbnail:
    image = Image.open(io.BytesIO(content))
    image.thumbnail((THUMBNAIL_SIZE, THUMBNAIL_SIZE))
    if image.mode in ("RGBA", "P"):
        image = image.convert("RGB")

    thumbnail = io.BytesIO()
    image.save(thumbnail, format="JPEG", optimize=True)
    return OpenGraphThumbnail.from_bytes(thumbnail.getvalue(), "image/jpeg")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import logging
from concurrent.futures import Future
from functools import partial
from pathlib import Path

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.const import IMAGE_MIME_TYPES
from gajim.common.const import VIDEO_MIME_TYPES
from gajim.common.multiprocess.thumbnail import create_thumbnail
from gajim.common.multiprocess.video_thumbnail import (
    extract_video_thumbnail_and_properties,
)
from gajim.common.util.filesystem import load_file_async
from gajim.common.util.image import image_size
from gajim.common.util.image import is_image_animated

from gajim.gtk.preview.animated_image import AnimatedImage
from gajim.gtk.preview.animated_image_backend import AnimatedImageBackend
from gajim.gtk.preview.animated_image_fallback_backend import (
    AnimatedImageFallbackBackend,
)
from gajim.gtk.preview.file_control_buttons import FileControlButtons
from gajim.gtk.preview.misc import LoadingBox  # noqa: F401 # type: ignore
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.preview.image")


MIN_PREVIEW_WIDTH = 100

# Previews are scaled down in steps of this many pixels.
PREVIEW_WIDTH_STEP = 10


class ImagePreviewLayout(Gtk.BinLayout):
    """Scales the preview down to the available width in discrete steps.

    do_measure() must report a height which does not depend on the width it is
    offered. A row which grows taller as it gets wider makes the height-for-width
    of the conversation's Gtk.ListBox non-monotonic, and GTK's size allocation
    then under-allocates it ("gtk_distribute_natural_allocation: assertion
    'extra_space >= 0' failed", triggered by an image preview followed by a
    wrapping message).

    The width the preview scales to is therefore not taken from the measured
    for_size, but re-picked from the last allocation, in steps and with
    hysteresis. Each individual measure pass stays self-consistent that way.
    """

    __gtype_name__ = "ImagePreviewLayout"

    def __init__(self) -> None:
        Gtk.BinLayout.__init__(self)
        self._dimension: tuple[int, int] | None = None
        self._preview_width: int | None = None
        self._update_id: int | None = None

    def set_preview_dimension(self, width: int, height: int) -> None:
        self._dimension = (max(width, 1), max(height, 1))
        self._preview_width = None
        self.layout_changed()

    def reset(self) -> None:
        if self._update_id is not None:
            GLib.source_remove(self._update_id)
            self._update_id = None

    def do_get_request_mode(self, widget: Gtk.Widget) -> Gtk.SizeRequestMode:
        return Gtk.SizeRequestMode.HEIGHT_FOR_WIDTH

    def _get_preview_size(self) -> tuple[int, int]:
        assert self._dimension is not None
        width, height = self._dimension

        if self._preview_width is None:
            # Nothing has been allocated yet, ask for the full size
            return width, height

        preview_width = min(
            max(self._preview_width, min(width, MIN_PREVIEW_WIDTH)), width
        )
        return preview_width, round(preview_width * height / width)

    def do_measure(
        self, widget: Gtk.Widget, orientation: Gtk.Orientation, for_size: int
    ) -> tuple[int, int, int, int]:
        if self._dimension is None:
            minimum, natural, *_ = Gtk.BinLayout.do_measure(
                self, widget, orientation, for_size
            )
            return minimum, natural, -1, -1

        width, _height = self._dimension
        _, preview_height = self._get_preview_size()

        if orientation == Gtk.Orientation.HORIZONTAL:
            # Keep asking for the full width, otherwise the preview could never
            # find out that there is room to grow again
            return min(width, MIN_PREVIEW_WIDTH), width, -1, -1

        return preview_height, preview_height, -1, -1

    def do_allocate(
        self, widget: Gtk.Widget, width: int, height: int, baseline: int
    ) -> None:
        available_width = width

        if self._dimension is not None:
            preview_width, _ = self._get_preview_size()
            # The height we reported belongs to preview_width, don't let the
            # image get any wider than that or it would be letterboxed
            width = min(width, preview_width)

        Gtk.BinLayout.do_allocate(self, widget, width, height, baseline)

        self._queue_preview_width(available_width)

    def _get_target_width(self, available_width: int) -> int:
        assert self._dimension is not None
        width, _height = self._dimension

        if available_width >= width:
            # There is room for the full size, don't round it down
            target = width
        else:
            target = max(
                available_width // PREVIEW_WIDTH_STEP * PREVIEW_WIDTH_STEP,
                PREVIEW_WIDTH_STEP,
            )

        if self._preview_width is None:
            return target

        if available_width < self._preview_width:
            # The preview does not fit anymore, it has to shrink
            return target

        if available_width - self._preview_width < 2 * PREVIEW_WIDTH_STEP:
            # Stay as is, only grow once there is a step to spare. This is the
            # hysteresis which keeps a scrollbar from toggling the size.
            return self._preview_width

        return target

    def _queue_preview_width(self, available_width: int) -> None:
        if self._dimension is None or self._update_id is not None:
            return

        target = self._get_target_width(available_width)
        if target == self._preview_width:
            return

        # Resizing from within size allocation is not allowed, defer it
        self._update_id = GLib.idle_add(self._apply_preview_width, target)

    def _apply_preview_width(self, target: int) -> bool:
        self._update_id = None

        if target != self._preview_width:
            self._preview_width = target
            self.layout_changed()

        return GLib.SOURCE_REMOVE


@Gtk.Template.from_string(string=get_ui_string("preview/image.ui"))
class ImagePreviewWidget(Gtk.Box, SignalManager):
    __gtype_name__ = "ImagePreviewWidget"

    __gsignals__ = {
        "display-error": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        )
    }

    _stack: Gtk.Stack = Gtk.Template.Child()
    _content_clamp: Adw.Clamp = Gtk.Template.Child()
    _content_overlay: Gtk.Overlay = Gtk.Template.Child()
    _image_button: Gtk.Button = Gtk.Template.Child()
    _picture: Gtk.Picture = Gtk.Template.Child()
    _play_image: Gtk.Image = Gtk.Template.Child()
    _file_control_buttons: FileControlButtons = Gtk.Template.Child()

    def __init__(
        self,
        filename: str,
        file_size: int,
        mime_type: str,
        orig_path: Path,
        thumb_path: Path,
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._orig_path = orig_path
        self._thumb_path = thumb_path
        self._filename = filename
        self._mime_type = mime_type

        self._layout = ImagePreviewLayout()
        self.set_layout_manager(self._layout)

        if mime_type in IMAGE_MIME_TYPES:
            self._type = "image"
        elif mime_type in VIDEO_MIME_TYPES:
            self._type = "video"
        else:
            raise ValueError("Not supported mime type: %s" % mime_type)

        content_hover_controller = Gtk.EventControllerMotion()
        self._connect(content_hover_controller, "enter", self._on_content_cursor_enter)
        self._connect(content_hover_controller, "leave", self._on_content_cursor_leave)
        self.add_controller(content_hover_controller)

        self._file_control_buttons.set_file_size(file_size)
        self._file_control_buttons.set_file_name(filename)
        self._file_control_buttons.set_path(orig_path)

        pointer_cursor = Gdk.Cursor.new_from_name("pointer")
        self._image_button.set_cursor(pointer_cursor)
        self._image_button.set_action_target_value(
            GLib.Variant("s", str(self._orig_path))
        )

        if not self._thumb_path.exists():
            self._create_thumbnail(self._type)
        else:
            load_file_async(self._thumb_path, self._on_thumb_load_finished)

    def _on_thumb_load_finished(
        self, data: bytes | None, error: GLib.Error | None, user_data: typing.Any
    ) -> None:
        if data is None:
            log.error("Loading thumbnail failed, %s: %s", self._thumb_path.name, error)
            self.emit("display-error")
            return

        self._thumbnail = data
        self._display_image_preview()

    def _create_thumbnail(self, type_: typing.Literal["image", "video"]) -> None:
        if type_ == "image":
            self._create_image_thumbnail()
        elif type_ == "video":
            self._create_video_thumbnail()

    def _create_image_thumbnail(self) -> None:
        assert self._thumb_path is not None
        assert self._orig_path is not None
        try:
            future = app.process_pool.submit(
                create_thumbnail,
                self._orig_path,
                self._thumb_path,
                app.settings.get("preview_size") * app.window.get_scale_factor(),
                self._mime_type,
            )
            future.add_done_callback(
                partial(GLib.idle_add, self._create_thumbnail_finished)
            )
        except Exception as error:
            log.warning("Creating thumbnail failed for: %s %s", self._orig_path, error)
            self.emit("display-error")

    def _create_video_thumbnail(self) -> None:
        assert self._orig_path is not None

        try:
            future = app.process_pool.submit(
                extract_video_thumbnail_and_properties,
                self._orig_path,
                self._thumb_path,
                app.settings.get("preview_size"),
            )
            future.add_done_callback(
                partial(GLib.idle_add, self._create_thumbnail_finished)
            )
        except Exception as error:
            log.warning("Creating thumbnail failed for: %s %s", self._orig_path, error)
            self.emit("display-error")

    def _create_thumbnail_finished(
        self, future: Future[tuple[bytes, dict[str, typing.Any]]]
    ) -> bool:
        try:
            thumbnail_bytes, _metadata = future.result()
        except Exception as error:
            log.exception(
                "Creating thumbnail failed for: %s %s", self._orig_path, error
            )
            self.emit("display-error")

        else:
            self._thumbnail = thumbnail_bytes
            self._display_image_preview()

        return GLib.SOURCE_REMOVE

    def _display_image_preview(self) -> None:
        if self._mime_type == "image/gif" and is_image_animated(self._orig_path):
            self._display_animated_image_preview(AnimatedImageBackend)
            return

        if self._mime_type in [
            "image/webp",
            "image/avif",
            "image/png",
        ] and is_image_animated(self._orig_path):
            self._display_animated_image_preview(AnimatedImageFallbackBackend)
            return

        self._display_static_image_preview()

    def _image_preview_dimension(
        self, image_width: int, image_height: int
    ) -> tuple[int, int]:
        max_preview_size = app.settings.get("preview_size")
        if image_width > max_preview_size or image_height > max_preview_size:
            # Scale down with or height to max_preview_size
            if image_width > image_height:
                width = max_preview_size
                height = int(max_preview_size / image_width * image_height)
            else:
                width = int(max_preview_size / image_height * image_width)
                height = max_preview_size
        else:
            width = image_width
            height = image_height

        return width, height

    def _set_preview_dimension(self, width: int, height: int) -> None:
        self._content_clamp.set_maximum_size(width)
        self._content_clamp.set_tightening_threshold(width)
        self._layout.set_preview_dimension(width, height)

    def _display_static_image_preview(self) -> None:
        try:
            texture = Gdk.Texture.new_from_bytes(GLib.Bytes.new(self._thumbnail))
        except GLib.Error:
            log.exception("Could not load image %s", self._filename)
            self.emit("display-error")
            return

        texture_width = texture.get_width()
        texture_height = texture.get_height()

        width, height = self._image_preview_dimension(texture_width, texture_height)

        self._set_preview_dimension(width, height)
        self._image_button.set_tooltip_text(self._filename)
        self._picture.set_paintable(texture)

        if self._type == "video":
            self._image_button.add_css_class("preview-video-overlay")
            self._play_image.set_pixel_size(min(width, height) // 3)
            self._play_image.set_visible(True)

        self._stack.set_visible_child_name("preview")

    def _display_animated_image_preview(
        self, backend: typing.Any[AnimatedImageBackend, AnimatedImageFallbackBackend]
    ) -> None:
        image_width, image_height = image_size(self._orig_path)
        width, height = self._image_preview_dimension(image_width, image_height)

        animated_image = AnimatedImage(self._thumb_path, self._orig_path, backend)
        pointer_cursor = Gdk.Cursor.new_from_name("pointer")
        animated_image.set_cursor(pointer_cursor)

        self._connect(animated_image, "error", self._on_animated_image_error)

        self._set_preview_dimension(width, height)
        self._content_overlay.set_child(animated_image)
        self._image_button.set_tooltip_text(self._filename)
        self._stack.set_visible_child_name("preview")

    def _on_animated_image_error(self, _animated_image: AnimatedImage) -> None:
        Gio.AppInfo.launch_default_for_uri(self._orig_path.as_uri(), None)

    def _on_content_cursor_enter(
        self,
        _controller: Gtk.EventControllerMotion,
        _x: int,
        _y: int,
    ) -> None:
        self._file_control_buttons.set_visible(True)

    def _on_content_cursor_leave(
        self,
        _controller: Gtk.EventControllerMotion,
    ) -> None:
        self._file_control_buttons.set_visible(False)

    def do_unroot(self) -> None:
        self._layout.reset()
        Gtk.Box.do_unroot(self)
        self._disconnect_all()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Generated by get_iana_data.py @ 2026-03-23T20:02:25.483403+00:00

URI_SCHEMES = {
    "aaa",
    "aaas",
    "about",
    "acap",
    "acct",
    "acd",
    "acr",
    "adiumxtra",
    "adt",
    "afp",
    "afs",
    "aim",
    "amss",
    "android",
    "appdata",
    "apt",
    "ar",
    "ari",
    "ark",
    "at",
    "attachment",
    "aw",
    "barion",
    "bb",
    "beshare",
    "bitcoin",
    "bitcoincash",
    "bl",
    "blob",
    "bluetooth",
    "bolo",
    "brid",
    "browserext",
    "cabal",
    "calculator",
    "callto",
    "cap",
    "caip",
    "cast",
    "casts",
    "chrome",
    "chrome-extension",
    "cid",
    "coap",
    "coap+tcp",
    "coap+ws",
    "coaps",
    "coaps+tcp",
    "coaps+ws",
    "com-eventbrite-attendee",
    "content",
    "content-type",
    "crid",
    "cstr",
    "cvs",
    "dab",
    "dat",
    "data",
    "dav",
    "dhttp",
    "diaspora",
    "dict",
    "did",
    "dis",
    "dlna-playcontainer",
    "dlna-playsingle",
    "dnp",
    "dns",
    "dntp",
    "doi",
    "donau",
    "dpp",
    "drm",
    "drop",
    "dtmi",
    "dtn",
    "dvb",
    "dvx",
    "dweb",
    "ed2k",
    "eid",
    "elsi",
    "embedded",
    "ens",
    "esim",
    "ethereum",
    "example",
    "facetime",
    "fax",
    "feed",
    "feedready",
    "fido",
    "file",
    "filesystem",
    "finger",
    "first-run-pen-experience",
    "fish",
    "fm",
    "ftp",
    "fuchsia-pkg",
    "geo",
    "gg",
    "git",
    "gitoid",
    "gizmoproject",
    "go",
    "gopher",
    "graph",
    "grd",
    "gtalk",
    "h323",
    "ham",
    "hcap",
    "hcp",
    "hs20",
    "http",
    "https",
    "hxxp",
    "hxxps",
    "hydrazone",
    "hyper",
    "i0",
    "iax",
    "ibi",
    "ibi-",
    "icap",
    "icon",
    "ilstring",
    "im",
    "imap",
    "info",
    "iotdisco",
    "ipfs",
    "ipn",
    "ipns",
    "ipp",
    "ipps",
    "irc",
    "irc6",
    "ircs",
    "iris",
    "iris.beep",
    "iris.lwz",
    "iris.xpc",
    "iris.xpcs",
    "isostore",
    "itms",
    "jabber",
    "jar",
    "jms",
    "keyparc",
    "lastfm",
    "lbry",
    "ldap",
    "ldaps",
    "leaptofrogans",
    "lid",
    "linkid",
    "lorawan",
    "lpa",
    "lvlt",
    "machineprovisioningprogressreporter",
    "magnet",
    "mailserver",
    "mailto",
    "maps",
    "market",
    "matrix",
    "message",
    "microsoft.windows.camera",
    "microsoft.windows.camera.multipicker",
    "microsoft.windows.camera.picker",
    "mid",
    "mms",
    "modem",
    "mongodb",
    "moz",
    "mqtt",
    "mqtts",
    "ms-access",
    "ms-appinstaller",
    "ms-browser-extension",
    "ms-calculator",
    "ms-drive-to",
    "ms-enrollment",
    "ms-excel",
    "ms-eyecontrolspeech",
    "ms-gamebarservices",
    "ms-gamingoverlay",
    "ms-getoffice",
    "ms-help",
    "ms-infopath",
    "ms-inputapp",
    "ms-launchremotedesktop",
    "ms-lockscreencomponent-config",
    "ms-media-stream-id",
    "ms-meetnow",
    "ms-mixedrealitycapture",
    "ms-mobileplans",
    "ms-newsandinterests",
    "ms-officeapp",
    "ms-people",
    "ms-personacard",
    "ms-powerpoint",
    "ms-project",
    "ms-publisher",
    "ms-recall",
    "ms-remotedesktop",
    "ms-remotedesktop-launch",
    "ms-restoretabcompanion",
    "ms-screenclip",
    "ms-screensketch",
    "ms-search",
    "ms-search-repair",
    "ms-secondary-screen-controller",
    "ms-secondary-screen-setup",
    "ms-settings",
    "ms-settings-airplanemode",
    "ms-settings-bluetooth",
    "ms-settings-camera",
    "ms-settings-cellular",
    "ms-settings-cloudstorage",
    "ms-settings-connectabledevices",
    "ms-settings-displays-topology",
    "ms-settings-emailandaccounts",
    "ms-settings-language",
    "ms-settings-location",
    "ms-settings-lock",
    "ms-settings-nfctransactions",
    "ms-settings-notifications",
    "ms-settings-power",
    "ms-settings-privacy",
    "ms-settings-proximity",
    "ms-settings-screenrotation",
    "ms-settings-wifi",
    "ms-settings-workplace",
    "ms-spd",
    "ms-stickers",
    "ms-sttoverlay",
    "ms-transit-to",
    "ms-useractivityset",
    "ms-uup",
    "ms-virtualtouchpad",
    "ms-visio",
    "ms-walk-to",
    "ms-whiteboard",
    "ms-whiteboard-cmd",
    "ms-widgetboard",
    "ms-widgets",
    "ms-word",
    "msnim",
    "msrp",
    "msrps",
    "mss",
    "mt",
    "mtqp",
    "mtrust",
    "mumble",
    "mupdate",
    "mvn",
    "mvrp",
    "mvrps",
    "news",
    "nfs",
    "ni",
    "nih",
    "nntp",
    "notes",
    "num",
    "ocf",
    "oid",
    "onenote",
    "onenote-cmd",
    "opaquelocktoken",
    "openid",
    "openpgp4fpr",
    "otpauth",
    "p1",
    "pack",
    "palm",
    "paparazzi",
    "payment",
    "payto",
    "pkcs11",
    "platform",
    "pop",
    "pres",
    "prospero",
    "proxy",
    "psyc",
    "pttp",
    "pwid",
    "qb",
    "query",
    "quic-transport",
    "redis",
    "rediss",
    "reload",
    "res",
    "resource",
    "rmi",
    "rsync",
    "rtmfp",
    "rtmp",
    "rtsp",
    "rtsps",
    "rtspu",
    "sarif",
    "secondlife",
    "secret-token",
    "service",
    "session",
    "sftp",
    "sgn",
    "shc",
    "shelter",
    "shttp",
    "sieve",
    "simpleledger",
    "simplex",
    "sip",
    "sips",
    "skype",
    "smb",
    "smp",
    "sms",
    "smtp",
    "snews",
    "snmp",
    "soap.beep",
    "soap.beeps",
    "soldat",
    "spacify",
    "spiffe",
    "spotify",
    "ssb",
    "ssh",
    "starknet",
    "steam",
    "stun",
    "stuns",
    "submit",
    "svn",
    "swh",
    "swid",
    "swidpath",
    "tag",
    "taler",
    "teamspeak",
    "teapot",
    "teapots",
    "tel",
    "teliaeid",
    "telnet",
    "tftp",
    "things",
    "thismessage",
    "thzp",
    "tip",
    "tn3270",
    "tool",
    "turn",
    "turns",
    "tv",
    "udp",
    "unreal",
    "upn",
    "upt",
    "urn",
    "ut2004",
    "uuid-in-package",
    "v-event",
    "vemmi",
    "ventrilo",
    "ves",
    "videotex",
    "view-source",
    "vnc",
    "vscode",
    "vscode-insiders",
    "vsls",
    "w3",
    "wais",
    "wasm",
    "wasm-js",
    "wcr",
    "web+ap",
    "web3",
    "webcal",
    "wifi",
    "wpid",
    "ws",
    "wss",
    "wtai",
    "wyciwyg",
    "xcompute",
    "xcon",
    "xcon-userid",
    "xfire",
    "xftp",
    "xmlrpc.beep",
    "xmlrpc.beeps",
    "xmpp",
    "xrcp",
    "xri",
    "ymsgr",
    "z39.50",
    "z39.50r",
    "z39.50s",
}



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import logging
import math
import sys
from pathlib import Path

import css_parser
from css_parser.css import CSSStyleRule
from css_parser.css import CSSStyleSheet
from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common import configpaths
from gajim.common.const import CSSPriority
from gajim.common.const import StyleAttr

if sys.platform == "win32":
    from gajim.common.winapi.system_style import SystemStyleListener
else:
    from gajim.common.dbus.system_style import SystemStyleListener

from gajim.gtk.const import Theme

log = logging.getLogger("gajim.gtk.css")


class CSSConfig:
    def __init__(self) -> None:
        """
        CSSConfig handles loading and storing of all relevant Gajim style files

        The order in which CSSConfig loads the styles

        1. gajim.css
        2. gajim-dark.css (Only if Adw.StyleManager.get_dark() = True)
        3. default.css or default-dark.css (from gajim/data/style)
        4. user-theme.css (from ~/.config/Gajim/theme)

        # gajim.css:

        This is the main style and the application default

        # gajim-dark.css

        Has only entries which we want to override in gajim.css

        # default.css or default-dark.css

        Has all the values that are changeable via UI (see themes.py).
        Depending on `Adw.StyleManager.get_dark()` either default.css or
        default-dark.css gets loaded

        # user-theme.css

        These are the themes the Themes Dialog stores. Because they are
        loaded at last they overwrite everything else. Users should add custom
        css here."""

        # Delete empty rules
        css_parser.ser.prefs.keepEmptyRules = False
        css_parser.ser.prefs.omitLastSemicolon = False

        # Holds the currently selected theme in the Theme Editor
        self._pre_css: CSSStyleSheet | None = None
        self._pre_css_path: Path | None = None

        # Holds the default theme, its used if values are not found
        # in the selected theme
        self._default_css: CSSStyleSheet | None = None
        self._default_css_path: Path | None = None

        # Holds the currently selected theme
        self._css: CSSStyleSheet | None = None
        self._css_path: Path | None = None

        # User Theme CSS Provider
        self._user_theme_provider = Gtk.CssProvider()

        self._default_providers: list[Gtk.CssProvider] = []

        # Used for dynamic classes like account colors
        self._dynamic_provider = Gtk.CssProvider()
        self._dynamic_dict: dict[str, str] = {}
        self.refresh()

        display = Gdk.Display.get_default()
        assert display is not None
        Gtk.StyleContext.add_provider_for_display(
            display, self._dynamic_provider, CSSPriority.APPLICATION
        )

        # Cache of recently requested values
        self._cache: dict[str, str | Pango.FontDescription | None] = {}

        # Holds all currently available themes
        self.themes: list[str] = []

        self._system_style = SystemStyleListener(callback=self.set_dark_theme)

        gtk_settings = Gtk.Settings.get_default()
        assert gtk_settings is not None
        self._gtk_settings = gtk_settings

        self.set_dark_theme()
        self._load_css()
        self._gather_available_themes()
        self._load_default()
        self._load_selected()
        self._activate_theme()

        Gtk.StyleContext.add_provider_for_display(
            display, self._user_theme_provider, CSSPriority.USER_THEME
        )

        self.apply_app_font_size()
        self._apply_platform_css()

    @property
    def prefer_dark(self) -> bool:
        setting = app.settings.get("dark_theme")
        if setting == Theme.SYSTEM:
            if self._system_style.prefer_dark is not None:
                return self._system_style.prefer_dark

            adw_style_manager = Adw.StyleManager.get_default()
            return adw_style_manager.get_dark()

        return setting == Theme.DARK

    def set_dark_theme(self, value: int | None = None) -> None:
        if value is None:
            value = app.settings.get("dark_theme")
        else:
            app.settings.set("dark_theme", value)

        adw_style_manager = Adw.StyleManager.get_default()
        if value == Theme.SYSTEM:
            if self._system_style.prefer_dark is None:
                adw_style_manager.set_color_scheme(Adw.ColorScheme.DEFAULT)
                return

            value = self._system_style.prefer_dark

        if value:
            adw_style_manager.set_color_scheme(Adw.ColorScheme.FORCE_DARK)
        else:
            adw_style_manager.set_color_scheme(Adw.ColorScheme.FORCE_LIGHT)

    def reload_css(self) -> None:
        display = Gdk.Display.get_default()
        assert display is not None

        for provider in self._default_providers:
            Gtk.StyleContext.remove_provider_for_display(display, provider)

        self._load_css()

    def _load_css(self) -> None:
        self._load_css_from_file("gajim.css", CSSPriority.APPLICATION)
        if self.prefer_dark:
            self._load_css_from_file("gajim-dark.css", CSSPriority.APPLICATION_DARK)

        self._load_css_from_file("default.css", CSSPriority.DEFAULT_THEME)
        if self.prefer_dark:
            self._load_css_from_file("default-dark.css", CSSPriority.DEFAULT_THEME_DARK)

    def _load_css_from_file(self, filename: str, priority: CSSPriority) -> None:
        path = configpaths.get("STYLE") / filename
        try:
            with open(path, encoding="utf8") as file_:
                css = file_.read()
        except Exception as exc:
            log.error("Error loading css: %s", exc)
            return
        self._activate_css(css, priority)

    def _activate_css(self, css: str, priority: CSSPriority) -> None:
        try:
            provider = Gtk.CssProvider()
            provider.load_from_bytes(GLib.Bytes.new(css.encode("utf-8")))
            self._default_providers.append(provider)

            display = Gdk.Display.get_default()
            assert display is not None
            Gtk.StyleContext.add_provider_for_display(display, provider, priority)
            self._load_selected()
            self._activate_theme()

        except Exception:
            log.exception("Error loading application css")

    @staticmethod
    def get_dpi_ratio() -> int:
        # Taken from https://gitlab.gnome.org/GNOME/gtk/-/blob/main/gtk/inspector/visual.c
        if sys.platform == "darwin":
            return 72 * 1024
        return 96 * 1024

    def apply_app_font_size(self) -> None:
        app_font_size = app.settings.get("app_font_size")
        if app_font_size == 1:
            log.debug("Reset gtk-xft-dpi")
            self._gtk_settings.reset_property("gtk-xft-dpi")
        else:
            dpi = app_font_size * self.get_dpi_ratio()
            log.debug("Set gtk-xft-dpi %s", dpi)
            self._gtk_settings.set_property("gtk-xft-dpi", dpi)

    def _apply_platform_css(self) -> None:
        """Apply extra CSS on Windows and macOS to fix issues, see:
        https://gitlab.gnome.org/GNOME/libadwaita/-/issues/1053
        """
        if sys.platform in ("win32", "darwin"):
            self._load_css_from_file(f"{sys.platform}.css", CSSPriority.APPLICATION)

    @staticmethod
    def _pango_to_css_weight(number: int) -> int:
        # Pango allows for weight values between 100 and 1000
        # CSS allows only full hundred numbers like 100, 200 ..
        if number < 100:
            return 100
        if number > 900:
            return 900
        return math.ceil(number / 100.0) * 100

    def _gather_available_themes(self) -> None:
        files = configpaths.get("MY_THEME").iterdir()
        self.themes = [file.stem for file in files if file.suffix == ".css"]
        if "default" in self.themes:
            # Ignore user created themes that are named 'default'
            self.themes.remove("default")

    def get_theme_path(self, theme: str, user: bool = True) -> Path:
        if theme == "default" and self.prefer_dark:
            theme = "default-dark"

        if user:
            return configpaths.get("MY_THEME") / f"{theme}.css"
        return configpaths.get("STYLE") / f"{theme}.css"

    def _determine_theme_path(self) -> Path:
        # Gets the path of the currently active theme.
        # If it does not exist, it falls back to the default theme
        theme = app.settings.get("roster_theme")
        if theme == "default":
            return self.get_theme_path(theme, user=False)

        theme_path = self.get_theme_path(theme)
        if not theme or not theme_path.exists():
            log.warning("Theme %s not found, fallback to default", theme)
            app.settings.set("roster_theme", "default")
            log.info("Use Theme: default")
            return self.get_theme_path("default", user=False)
        log.info("Use Theme: %s", theme)
        return theme_path

    def _load_selected(self, new_path: Path | None = None) -> None:
        if new_path is None:
            self._css_path = self._determine_theme_path()
        else:
            self._css_path = new_path
        self._css = css_parser.parseFile(self._css_path)

    def _load_default(self) -> None:
        self._default_css_path = self.get_theme_path("default", user=False)
        self._default_css = css_parser.parseFile(self._default_css_path)

    def _load_pre(self, theme: str) -> None:
        log.info("Preload theme %s", theme)
        self._pre_css_path = self.get_theme_path(theme)
        self._pre_css = css_parser.parseFile(self._pre_css_path)

    def _write(self, pre: bool) -> None:
        path = self._css_path
        css = self._css
        if pre:
            path = self._pre_css_path
            css = self._pre_css

        assert path is not None
        assert css is not None
        with open(path, "w", encoding="utf-8") as file:
            file.write(css.cssText.decode("utf-8"))

        active = self._pre_css_path == self._css_path
        if not pre or active:
            self._load_selected()
            self._activate_theme()

    def set_value(
        self,
        selector: str,
        attr: str | StyleAttr,
        value: str | Pango.FontDescription,
        pre: bool = False,
    ) -> None:
        if attr == StyleAttr.FONT:
            # forward to set_font() for convenience
            assert isinstance(value, Pango.FontDescription)
            self.set_font(selector, value, pre)
            return

        if isinstance(attr, StyleAttr):
            attr = attr.value

        css = self._css
        if pre:
            css = self._pre_css

        assert css is not None
        for rule in css:
            if rule.type != rule.STYLE_RULE:
                continue
            if rule.selectorText == selector:
                log.info("Set %s %s %s", selector, attr, value)
                rule.style[attr] = value
                if not pre:
                    self._add_to_cache(selector, attr, value)
                self._write(pre)
                return

        # The rule was not found, so we add it to this theme
        log.info("Set %s %s %s", selector, attr, value)
        rule = CSSStyleRule(selectorText=selector)
        rule.style[attr] = value
        css.add(rule)
        self._write(pre)

    def set_font(
        self, selector: str, description: Pango.FontDescription, pre: bool = False
    ) -> None:
        css = self._css
        if pre:
            css = self._pre_css
        family, size, style, weight = self._get_attr_from_description(description)
        assert css is not None
        for rule in css:
            if rule.type != rule.STYLE_RULE:
                continue
            if rule.selectorText == selector:
                log.info(
                    "Set Font for: %s %s %s %s %s",
                    selector,
                    family,
                    size,
                    style,
                    weight,
                )
                # Quote font-family in order to avoid unquoted font-families
                # (this is a bug in css_parser)
                rule.style["font-family"] = f'"{family}"'
                rule.style["font-style"] = style
                rule.style["font-size"] = f"{size}pt"
                rule.style["font-weight"] = weight

                if not pre:
                    self._add_to_cache(selector, "fontdescription", description)
                self._write(pre)
                return

        # The rule was not found, so we add it to this theme
        log.info("Set Font for: %s %s %s %s %s", selector, family, size, style, weight)
        rule = CSSStyleRule(selectorText=selector)
        # Quote font-family in order to avoid unquoted font-families
        # (this is a bug in css_parser)
        rule.style["font-family"] = f'"{family}"'
        rule.style["font-style"] = style
        rule.style["font-size"] = f"{size}pt"
        rule.style["font-weight"] = weight
        css.add(rule)
        self._write(pre)

    def _get_attr_from_description(
        self, description: Pango.FontDescription
    ) -> tuple[str | None, float, str, int]:
        size = description.get_size() / Pango.SCALE
        style = self._get_string_from_pango_style(description.get_style())
        weight = self._pango_to_css_weight(int(description.get_weight()))
        family = description.get_family()
        return family, size, style, weight

    def _get_default_rule(self, selector: str, _attr: str) -> CSSStyleRule | None:
        assert self._default_css is not None
        for rule in self._default_css:
            if rule.type != rule.STYLE_RULE:
                continue
            if rule.selectorText == selector:
                log.info("Get Default Rule %s", selector)
                return rule
        return None

    def get_font(
        self, selector: str, pre: bool = False
    ) -> Pango.FontDescription | None:
        if pre:
            css = self._pre_css
        else:
            css = self._css
            try:
                font_desc = self._get_from_cache(selector, "fontdescription")
                if isinstance(font_desc, Pango.FontDescription):
                    return font_desc
            except KeyError:
                pass

        if css is None:
            return None

        for rule in css:
            if rule.type != rule.STYLE_RULE:
                continue
            if rule.selectorText == selector:
                log.info("Get Font for: %s", selector)
                style = rule.style.getPropertyValue("font-style") or None
                size = rule.style.getPropertyValue("font-size") or None
                weight = rule.style.getPropertyValue("font-weight") or None
                family = rule.style.getPropertyValue("font-family") or None

                if family is not None:
                    # Unquote previously quoted font-family
                    # (this is a bug in css_parser)
                    family = family.strip('"')

                desc = self._get_description_from_css(family, size, style, weight)
                if not pre:
                    self._add_to_cache(selector, "fontdescription", desc)
                return desc

        self._add_to_cache(selector, "fontdescription", None)
        return None

    def _get_description_from_css(
        self,
        family: str | None,
        size: str | None,
        style: str | None,
        weight: str | None,
    ) -> Pango.FontDescription | None:
        if family is None:
            return None
        desc = Pango.FontDescription()
        desc.set_family(family)
        if weight is not None:
            desc.set_weight(Pango.Weight(int(weight)))
        if style is not None:
            desc.set_style(self._get_pango_style_from_string(style))
        if size is not None:
            desc.set_size(int(float(size[:-2])) * Pango.SCALE)
        return desc

    @staticmethod
    def _get_pango_style_from_string(style: str) -> Pango.Style:
        if style == "normal":
            return Pango.Style(0)
        if style == "oblique":
            return Pango.Style(1)
        # Pango.Style.ITALIC:
        return Pango.Style(2)

    @staticmethod
    def _get_string_from_pango_style(style: Pango.Style) -> str:
        if style == Pango.Style.NORMAL:
            return "normal"
        if style == Pango.Style.ITALIC:
            return "italic"
        # Pango.Style.OBLIQUE:
        return "oblique"

    def get_value(
        self, selector: str, attr: str | StyleAttr, pre: bool = False
    ) -> str | Pango.FontDescription | None:
        if attr == StyleAttr.FONT:
            # forward to get_font() for convenience
            return self.get_font(selector, pre)

        if isinstance(attr, StyleAttr):
            attr = attr.value

        if pre:
            css = self._pre_css
        else:
            css = self._css
            try:
                return self._get_from_cache(selector, attr)
            except KeyError:
                pass

        if css is not None:
            for rule in css:
                if rule.type != rule.STYLE_RULE:
                    continue
                if rule.selectorText == selector:
                    log.info("Get %s %s: %s", selector, attr, rule.style[attr] or None)
                    value = rule.style.getPropertyValue(attr) or None
                    if not pre:
                        self._add_to_cache(selector, attr, value)
                    return value

        # We didn’t find the selector in the selected theme
        # search in default theme
        if not pre:
            rule = self._get_default_rule(selector, attr)
            value = rule if rule is None else rule.style[attr]
            self._add_to_cache(selector, attr, value)
            return value
        return None

    def remove_value(
        self, selector: str, attr: str | StyleAttr, pre: bool = False
    ) -> None:
        if attr == StyleAttr.FONT:
            # forward to remove_font() for convenience
            self.remove_font(selector, pre)
            return

        if isinstance(attr, StyleAttr):
            attr = attr.value

        css = self._css
        if pre:
            css = self._pre_css

        assert css is not None
        for rule in css:
            if rule.type != rule.STYLE_RULE:
                continue
            if rule.selectorText == selector:
                log.info("Remove %s %s", selector, attr)
                rule.style.removeProperty(attr)
                break
        self._write(pre)

    def remove_font(self, selector: str, pre: bool = False) -> None:
        css = self._css
        if pre:
            css = self._pre_css

        assert css is not None
        for rule in css:
            if rule.type != rule.STYLE_RULE:
                continue
            if rule.selectorText == selector:
                log.info("Remove Font from: %s", selector)
                rule.style.removeProperty("font-family")
                rule.style.removeProperty("font-size")
                rule.style.removeProperty("font-style")
                rule.style.removeProperty("font-weight")
                break
        self._write(pre)

    def change_theme(self, theme: str) -> bool:
        user = not theme == "default"
        theme_path = self.get_theme_path(theme, user=user)
        if not theme_path.exists():
            log.error("Change Theme: Theme %s does not exist", theme_path)
            return False
        self._load_selected(theme_path)
        self._activate_theme()
        app.settings.set("roster_theme", theme)
        log.info("Change Theme: Successful switched to %s", theme)
        return True

    def change_preload_theme(self, theme: str) -> bool:
        theme_path = self.get_theme_path(theme)
        if not theme_path.exists():
            log.error("Change Preload Theme: Theme %s does not exist", theme_path)
            return False
        self._load_pre(theme)
        log.info("Successful switched to %s", theme)
        return True

    def rename_theme(self, old_theme: str, new_theme: str) -> bool:
        if old_theme not in self.themes:
            log.error("Rename Theme: Old theme %s not found", old_theme)
            return False

        if new_theme in self.themes:
            log.error("Rename Theme: New theme %s exists already", new_theme)
            return False

        old_theme_path = self.get_theme_path(old_theme)
        new_theme_path = self.get_theme_path(new_theme)
        old_theme_path.rename(new_theme_path)
        self.themes.remove(old_theme)
        self.themes.append(new_theme)
        self._load_pre(new_theme)
        log.info(
            "Rename Theme: Successful renamed theme from %s to %s", old_theme, new_theme
        )
        return True

    def _activate_theme(self) -> None:
        log.info("Activate theme")
        self._invalidate_cache()
        assert self._css is not None
        self._user_theme_provider.load_from_bytes(GLib.Bytes.new(self._css.cssText))

    def add_new_theme(self, theme: str) -> bool:
        theme_path = self.get_theme_path(theme)
        if theme_path.exists():
            log.error("Add Theme: %s exists already", theme_path)
            return False
        with open(theme_path, "w", encoding="utf8"):
            pass
        self.themes.append(theme)
        log.info("Add Theme: Successful added theme %s", theme)
        return True

    def remove_theme(self, theme: str) -> None:
        theme_path = self.get_theme_path(theme)
        if theme_path.exists():
            theme_path.unlink()
            self.themes.remove(theme)
        log.info("Remove Theme: Successful removed theme %s", theme)

    def _add_to_cache(
        self, selector: str, attr: str, value: str | Pango.FontDescription | None
    ) -> None:
        self._cache[selector + attr] = value

    def _get_from_cache(
        self, selector: str, attr: str
    ) -> str | Pango.FontDescription | None:
        return self._cache[selector + attr]

    def _invalidate_cache(self) -> None:
        self._cache = {}

    def refresh(self) -> None:
        css = ""
        accounts = app.settings.get_accounts()
        for index, account in enumerate(accounts):
            color = app.settings.get_account_setting(account, "account_color")
            css_class = f"gajim_class_{index}"
            css += f".{css_class} {{ background-color: {color}; }}\n"
            self._dynamic_dict[account] = css_class

        self._dynamic_provider.load_from_bytes(GLib.Bytes.new(css.encode()))

    def get_dynamic_class(self, name: str) -> str:
        return self._dynamic_dict[name]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import TYPE_CHECKING

import weakref
from datetime import datetime

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.const import TRUST_SYMBOL_DATA
from gajim.common.i18n import _
from gajim.common.i18n import p_
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.storage.archive.const import MessageState
from gajim.common.types import ChatContactT

from gajim.gtk.emoji_chooser import EmojiChooser
from gajim.gtk.menus import get_groupchat_participant_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.widgets import GajimPopover

if TYPE_CHECKING:
    from gajim.gtk.conversation.rows.message import MessageRow


class SimpleLabel(Gtk.Label):
    def __init__(self) -> None:
        Gtk.Label.__init__(self)
        self.set_selectable(True)
        self.set_xalign(0)
        self.set_wrap(True)
        self.set_wrap_mode(Pango.WrapMode.WORD_CHAR)


class MessageRowActions(Gtk.Box):
    _FADE_IN_DURATION = 150

    def __init__(self) -> None:
        Gtk.Box.__init__(
            self,
            halign=Gtk.Align.END,
            valign=Gtk.Align.START,
            margin_start=40,
            margin_end=40,
            visible=False,
        )

        self._message_row: weakref.ReferenceType[MessageRow] | None = None
        self._contact: ChatContactT | None = None

        self._has_cursor = False
        self._is_menu_open = False
        self._is_scrolling = False

        self._timeout_id: int | None = None

        self._fade_in_animation = Adw.TimedAnimation(
            widget=self,
            value_from=0,
            value_to=1,
            duration=self._FADE_IN_DURATION,
            target=Adw.PropertyAnimationTarget.new(self, "opacity"),
        )

        self._reaction_buttons: list[Gtk.Button | Gtk.MenuButton] = []

        for emoji in app.settings.get("quick_reaction_emojis").split(","):
            button = QuickReactionButton(emoji)
            button.connect("clicked", self._on_quick_reaction_button_clicked)
            self._reaction_buttons.append(button)

        choose_reaction_button = Gtk.MenuButton(
            icon_name="lucide-smile-plus-symbolic",
            tooltip_text=_("Add Reaction…"),
            visible=False,
        )
        choose_reaction_button.set_create_popup_func(self._on_emoji_create_popover)

        self._reaction_buttons.append(choose_reaction_button)

        self._reply_button = Gtk.Button.new_from_icon_name("lucide-reply-symbolic")
        self._reply_button.set_visible(False)
        self._reply_button.set_tooltip_text(_("Reply…"))
        self._reply_button.connect("clicked", self._on_reply_clicked)

        self._more_popover = Gtk.PopoverMenu()
        self._more_popover.connect("closed", self._on_popover_closed)

        self._more_button = Gtk.MenuButton(
            icon_name="lucide-ellipsis-symbolic", popover=self._more_popover
        )
        self._more_button.set_create_popup_func(self._on_create_more_popover)

        box = Gtk.Box()
        box.add_css_class("linked")
        box.add_css_class("row-actions-button-box")

        for button in self._reaction_buttons:
            box.append(button)

        box.append(self._reply_button)
        box.append(self._more_button)

        self.append(box)

        hover_controller = Gtk.EventControllerMotion()
        hover_controller.connect("enter", self._on_cursor_enter)
        hover_controller.connect("leave", self._on_cursor_leave)
        self.add_controller(hover_controller)

    def hide_actions(self) -> None:
        if self._is_menu_open:
            return

        # Set a 10ms timeout, which gives us enough time to inhibit hiding
        # if the cursor enters (cursor changes from row to MessageRowActions)
        self._hide_with_timeout()

    def set_scrolling(self, scrolling: bool) -> None:
        if self._is_scrolling == scrolling:
            return

        self._is_scrolling = scrolling

        if self._is_menu_open:
            return

        if scrolling:
            self._fade_in_animation.reset()
            self.set_opacity(0)
        else:
            self._fade_in_animation.play()

    def get_message_row(self) -> MessageRow | None:
        if self._message_row is None:
            return None
        return self._message_row()

    def update(self, y_coord: float, message_row: MessageRow) -> None:
        if self._is_menu_open:
            return

        self._message_row = weakref.ref(message_row)

        if self._timeout_id is not None:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = None

        self_height = self.get_height()
        y_coord = max(y_coord, self_height)

        # Subtract some space to let MessageRowActions 'flow' above the row,
        # but make sure to make offset smaller than row height.
        offset = 12
        if message_row.is_merged:
            offset = min(28, self_height - 6)

        adjusted_y_coord = y_coord - offset
        adjusted_y_coord = max(adjusted_y_coord, 0)

        self.set_margin_top(int(adjusted_y_coord))

        message_row_width = message_row.get_width()
        reactions_visible = message_row.can_react()

        for button in self._reaction_buttons:
            if (
                isinstance(button, QuickReactionButton)
                and reactions_visible
                and message_row_width < 600
            ):
                # Don't show QuickReactionButtons on narrow screens
                button.set_visible(False)
                continue

            button.set_visible(reactions_visible)

        self._reply_button.set_visible(message_row.can_reply())
        self.set_visible(True)

    def switch_contact(self, contact: ChatContactT) -> None:
        self._message_row = None
        self._contact = contact
        self._is_scrolling = False
        self._fade_in_animation.reset()
        self.set_opacity(1)
        self.set_visible(False)

    def _hide_with_timeout(self) -> None:
        if self._timeout_id is not None:
            GLib.source_remove(self._timeout_id)

        self._timeout_id = GLib.timeout_add(10, self._hide)

    def _hide(self) -> None:
        self._timeout_id = None

        if self._has_cursor or self._message_row is None:
            return

        if message_row := self.get_message_row():
            message_row.remove_css_class("conversation-row-hover")

        self.set_visible(False)

    def _on_cursor_enter(
        self,
        controller: Gtk.EventControllerMotion,
        _x: int,
        _y: int,
    ) -> None:
        self._has_cursor = True

        if self._is_menu_open:
            return

        if message_row := self.get_message_row():
            # message_row may be None if MessageRowActions are entered without
            # hovering a MessageRow before (e.g. by switching chats via shortcut)
            message_row.add_css_class("conversation-row-hover")

    def _on_cursor_leave(self, controller: Gtk.EventControllerMotion) -> None:
        self._has_cursor = False

        if self._is_menu_open:
            # A popover triggers a leave event,
            # but we don't want to hide MessageRowActions
            return

        if message_row := self.get_message_row():
            message_row.remove_css_class("conversation-row-hover")

    def _on_reply_clicked(self, _button: Gtk.Button) -> None:
        if message_row := self.get_message_row():
            app.window.activate_action("win.reply", GLib.Variant("u", message_row.pk))

    def _on_quick_reaction_button_clicked(self, button: QuickReactionButton) -> None:
        self._send_reaction(button.emoji)

    def _on_emoji_create_popover(self, button: Gtk.MenuButton) -> None:
        self._is_menu_open = True
        emoji_chooser = app.window.get_emoji_chooser()
        button.set_popover(emoji_chooser)
        emoji_chooser.set_emoji_picked_func(self._on_reaction_added)
        emoji_chooser.connect_after("closed", self._on_popover_closed)

    def _on_reaction_added(self, _widget: EmojiChooser, emoji: str) -> None:
        # Remove emoji variant selectors
        emoji = emoji.strip("\ufe0e\ufe0f")
        self._send_reaction(emoji, toggle=False)

    def _send_reaction(self, emoji: str, toggle: bool = True) -> None:
        if message_row := self.get_message_row():
            message_row.send_reaction(emoji, toggle)

    def _on_create_more_popover(self, button: Gtk.MenuButton) -> None:
        message_row = self.get_message_row()
        if message_row is None:
            return

        self._is_menu_open = True

        menu = message_row.get_chat_row_menu()
        self._more_popover.set_menu_model(menu)

    def _on_popover_closed(self, popover: Gtk.PopoverMenu) -> None:
        if isinstance(popover, EmojiChooser):
            popover.disconnect_by_func(self._on_popover_closed)

        if message_row := self.get_message_row():
            message_row.remove_css_class("conversation-row-hover")

        self._is_menu_open = False
        self._hide_with_timeout()


class QuickReactionButton(Gtk.Button):
    def __init__(self, emoji: str) -> None:
        self.emoji = emoji

        # Add emoji presentation selector, otherwise depending on the font
        # emojis might be displayed in its text variant
        emoji_presentation_form = f"{emoji}\ufe0f"

        Gtk.Button.__init__(
            self,
            label=emoji_presentation_form,
            tooltip_text=_("React with %s") % emoji_presentation_form,
            visible=False,
        )

        self.remove_css_class("text-button")
        self.add_css_class("image-button")


class DateTimeLabel(Gtk.Label):
    def __init__(self, timestamp: datetime) -> None:
        Gtk.Label.__init__(self)

        format_string = app.settings.get("time_format")
        if timestamp.date() < datetime.today().date():
            format_string = app.settings.get("date_time_format")

        self.set_text(timestamp.strftime(format_string))
        self.set_selectable(True)
        self.set_halign(Gtk.Align.START)
        self.set_valign(Gtk.Align.END)
        self.set_margin_start(6)
        self.set_margin_end(3)
        self.add_css_class("conversation-meta")
        format_string = app.settings.get("date_time_format")
        self.set_tooltip_text(timestamp.strftime(format_string))


class NicknameLabel(Gtk.Label):
    def __init__(self, name: str, from_us: bool) -> None:
        Gtk.Label.__init__(
            self,
            selectable=True,
            ellipsize=Pango.EllipsizeMode.END,
            max_width_chars=32,
            valign=Gtk.Align.END,
            label=name,
        )
        self.add_css_class("conversation-nickname")

        if from_us:
            css_class = "gajim-outgoing-nickname"
        else:
            css_class = "gajim-incoming-nickname"

        self.add_css_class(css_class)


class MessageIcons(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=3)

        self._encryption_image = Gtk.Image()
        self._encryption_image.set_visible(False)
        self._encryption_image.set_margin_end(6)

        self._security_label = Gtk.Label()
        self._security_label.set_visible(False)
        self._security_label.set_margin_end(6)
        self._security_label.set_ellipsize(Pango.EllipsizeMode.END)
        self._security_label.set_max_width_chars(20)

        self._correction_image = Gtk.Image.new_from_icon_name(
            "lucide-square-pen-symbolic"
        )
        self._correction_image.set_visible(False)
        self._correction_image.add_css_class("dimmed")

        self._message_state_image = Gtk.Image()
        self._message_state_image.set_visible(False)
        self._message_state_image.add_css_class("dimmed")

        self._marker_image = Gtk.Image.new_from_icon_name("lucide-check-symbolic")
        self._marker_image.set_visible(False)
        self._marker_image.add_css_class("dimmed")
        self._marker_image.set_tooltip_text(p_("Message state", "Received"))

        self._error_image = Gtk.Image.new_from_icon_name("lucide-circle-alert-symbolic")
        self._error_image.add_css_class("warning")
        self._error_image.set_visible(False)

        self._mention_image = Gtk.Image.new_from_icon_name("lucide-at-sign-symbolic")
        self._mention_image.add_css_class("conversation-mention-fg-color")
        self._mention_image.set_tooltip_text(_("You have been mentioned"))
        self._mention_image.set_visible(False)

        self.append(self._encryption_image)
        self.append(self._security_label)
        self.append(self._correction_image)
        self.append(self._message_state_image)
        self.append(self._marker_image)
        self.append(self._error_image)
        self.append(self._mention_image)

    def set_encryption_icon_visible(self, visible: bool) -> None:
        self._encryption_image.set_visible(visible)

    def set_encrytion_icon_data(self, icon: str, color: str, tooltip: str) -> None:
        for trust_data in TRUST_SYMBOL_DATA.values():
            self._encryption_image.remove_css_class(trust_data[2])

        self._encryption_image.add_css_class(color)
        self._encryption_image.set_from_icon_name(icon)
        self._encryption_image.set_tooltip_markup(tooltip)

    def set_security_label_visible(self, visible: bool) -> None:
        self._security_label.set_visible(visible)

    def set_security_label_data(self, tooltip: str, markup: str) -> None:
        self._security_label.set_tooltip_text(tooltip)
        self._security_label.set_markup(markup)

    def set_receipt_icon_visible(self, visible: bool) -> None:
        if not app.settings.get("positive_184_ack"):
            return
        self._marker_image.set_visible(visible)

    def set_message_state_icon(self, state: MessageState) -> None:
        if state == MessageState.PENDING:
            icon_name = "lucide-clock-symbolic"
            tooltip_text = _("Pending")
        else:
            icon_name = "lucide-check-symbolic"
            tooltip_text = _("Received")
        self._message_state_image.set_from_icon_name(icon_name)
        self._message_state_image.set_tooltip_text(tooltip_text)
        self._message_state_image.set_visible(True)

    def hide_message_state_icon(self):
        self._message_state_image.set_visible(False)

    def set_correction_icon_visible(self, visible: bool) -> None:
        self._correction_image.set_visible(visible)

    def set_correction_tooltip(self, text: str) -> None:
        self._correction_image.set_tooltip_text(text)

    def set_error_icon_visible(self, visible: bool) -> None:
        self._error_image.set_visible(visible)

    def set_error_tooltip(self, text: str) -> None:
        self._error_image.set_tooltip_text(text)

    def set_mention_icon_visibe(self, visible: bool) -> None:
        self._mention_image.set_visible(visible)


class AvatarBox(Gtk.Box, SignalManager):
    def __init__(self, contact: ChatContactT) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self.set_size_request(AvatarSize.ROSTER, -1)
        self.set_valign(Gtk.Align.START)

        self._contact = contact
        self._name = ""

        self._image = Gtk.Image(pixel_size=AvatarSize.ROSTER)
        self.append(self._image)

        if self._contact.is_groupchat:
            self.set_cursor(Gdk.Cursor.new_from_name("pointer"))

        self._menu_popover = GajimPopover(None)
        self.append(self._menu_popover)

        gesture_left_click = Gtk.GestureClick(button=Gdk.BUTTON_PRIMARY)
        self._connect(gesture_left_click, "pressed", self._on_avatar_clicked)
        self.add_controller(gesture_left_click)

        gesture_right_click = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(gesture_right_click, "pressed", self._on_avatar_clicked)
        self.add_controller(gesture_right_click)

    def do_unroot(self):
        self._disconnect_all()
        Gtk.Box.do_unroot(self)

    def set_from_paintable(self, texture: Gdk.Texture | None) -> None:
        self._image.set_from_paintable(texture)

    def set_name(self, name: str) -> None:
        self._name = name

    def set_merged(self, merged: bool) -> None:
        self._image.set_visible(not merged)

    def _on_avatar_clicked(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        if not isinstance(self._contact, GroupchatContact):
            return

        if gesture_click.get_current_button() == Gdk.BUTTON_PRIMARY:
            app.window.activate_action(
                "win.input-mention", GLib.Variant("s", self._name)
            )
        elif gesture_click.get_current_button() == Gdk.BUTTON_SECONDARY:
            self._show_participant_menu(self._name, x, y)

    def _show_participant_menu(self, nick: str, x: float, y: float) -> None:
        assert isinstance(self._contact, GroupchatContact)
        if not self._contact.is_joined:
            return

        self_contact = self._contact.get_self()
        assert self_contact is not None

        if nick == self_contact.name:
            # Don’t show menu for us
            return

        contact = self._contact.get_resource(nick)
        menu = get_groupchat_participant_menu(
            self._contact.account, self_contact, contact
        )

        self._menu_popover.set_menu_model(menu)
        self._menu_popover.set_pointing_to_coord(x, y)
        self._menu_popover.popup()


class MentionIndicator(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(self, visible=False)
        self.add_css_class("conversation-mention-bg-color")
        self.add_css_class("ps-3")
        self.add_css_class("card")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0054: vcard-temp

from __future__ import annotations

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import DiscoInfo

from gajim.common import types
from gajim.common.modules.base import BaseModule


class VCardTemp(BaseModule):
    _nbxmpp_extends = "VCardTemp"
    _nbxmpp_methods = [
        "request_vcard",
        "set_vcard",
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self._own_vcard = None
        self.supported = False

    def pass_disco(self, info: DiscoInfo) -> None:
        if Namespace.VCARD not in info.features:
            return

        self.supported = True
        self._log.info("Discovered vcard-temp: %s", info.jid)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.events import MucInvitation
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.util.muc import get_group_chat_nick

from gajim.gtk.groupchat_info import GroupChatInfoScrolled
from gajim.gtk.groupchat_nick_chooser import GroupChatNickChooser
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.widgets import AccountBadge


@Gtk.Template(string=get_ui_string("groupchat_invitation.ui"))
class GroupChatInvitation(Gtk.Box, SignalManager):
    __gtype_name__ = "GroupChatInvitation"

    _inviter_avatar: Gtk.Image = Gtk.Template.Child()
    _inviter_name_label: Gtk.Label = Gtk.Template.Child()
    _account_badge: AccountBadge = Gtk.Template.Child()

    _info_box: Gtk.Box = Gtk.Template.Child()

    _actions_box: Gtk.Box = Gtk.Template.Child()
    _decline_button: Gtk.Button = Gtk.Template.Child()
    _join_button: Gtk.Button = Gtk.Template.Child()
    _nick_chooser: GroupChatNickChooser = Gtk.Template.Child()

    __gsignals__ = {
        "accepted": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        ),
        "declined": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        ),
    }

    def __init__(self, account: str, event: MucInvitation) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._client = app.get_client(account)
        self._room_jid = str(event.muc)
        self._from = str(event.from_)
        self._password = event.password

        muc_info_box = GroupChatInfoScrolled(account, minimal=True)
        muc_info_box.set_hexpand(True)
        muc_info_box.set_from_disco_info(event.info)
        self._info_box.append(muc_info_box)

        contact = self._client.get_module("Contacts").get_contact(event.from_.bare)
        assert isinstance(contact, BareContact | GroupchatContact)
        self._inviter_avatar.set_from_paintable(
            contact.get_avatar(AvatarSize.SMALL, self.get_scale_factor())
        )
        self._inviter_name_label.set_text(contact.name)

        enabled_accounts = app.get_enabled_accounts_with_labels()
        self._account_badge.set_account(account)
        self._account_badge.set_visible(len(enabled_accounts) > 1)

        self._connect(self._decline_button, "clicked", self._on_decline)
        self._connect(self._join_button, "clicked", self._on_join)

        self._nick_chooser.set_text(get_group_chat_nick(self._account, event.info.jid))

        self._join_button.grab_focus()

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def _on_join(self, _button: Gtk.Button) -> None:
        nickname = self._nick_chooser.get_text()
        app.window.show_add_join_groupchat(
            self._account, self._room_jid, nickname=nickname
        )
        self.emit("accepted")

    def _on_decline(self, _button: Gtk.Button) -> None:
        self._client.get_module("MUC").decline(self._room_jid, self._from)
        self.emit("declined")

    def disable_buttons(self) -> None:
        self._actions_box.set_sensitive(False)
